import os
from pkg_resources import resource_filename

from nive.components.reform.exception import TemplateError
from translationstring import ChameleonTranslate

try: # pragma: no cover (chameleon 1)
    from chameleon.zpt import language # will raise exception for c2
    from chameleon.zpt.template import PageTemplateFile

    def cache(func):
        def load(self, *args):
            template = self.registry.get(args)
            if template is None:
                self.registry[args] = template = func(self, *args)
            return template
        return load

    class ZPTTemplateLoader(object):
        """ A Chameleon ZPT template loader """
        parser = language.Parser()

        def __init__(self, search_path=None, auto_reload=True, debug=True,
                     encoding='utf-8', translate=None):
            if search_path is None:
                search_path = []
            if isinstance(search_path, str):
                search_path = [search_path]
            self.search_path = search_path
            self.auto_reload = auto_reload
            self.debug = debug
            self.encoding = encoding
            self.translate = translate
            self.registry = {}
            self.notexists = {}

        @cache
        def load(self, filename):
            for path in self.search_path:
                path = os.path.join(path, filename)
                if (path in self.notexists) and (not self.auto_reload):
                    raise TemplateError("Can not find template %s" % filename)
                try:
                    return PageTemplateFile(path, parser=self.parser,
                                            auto_reload=self.auto_reload,
                                            debug = self.debug,
                                            encoding=self.encoding,
                                            translate=self.translate)
                except OSError:
                    self.notexists[path] = True

            raise TemplateError("Can not find template %s" % filename)
except ImportError:
    # Chameleon 2
    from chameleon.zpt.loader import TemplateLoader
    class ZPTTemplateLoader(TemplateLoader):
        def __init__(self, *args, **kwargs):
            kwargs.setdefault('encoding', 'utf-8')
            super(ZPTTemplateLoader, self).__init__(*args, **kwargs)

        def load(self, filename, *args, **kwargs):
            try:
                return super(ZPTTemplateLoader, self).load(
                    filename, *args, **kwargs)
            except ValueError:
                raise TemplateError(filename)

class ZPTRendererFactory(object):
    """
    Construct a Chameleon ZPT :term:`renderer`.

    The resulting renderer callable accepts a template name *without*
    the ``.pt`` file extension and a set of keyword` values.

    **Arguments**

    search_path
      A sequence of strings representing fully qualified filesystem
      directories containing reform Chameleon template sources.  The
      order in which the directories are listed within ``search_path``
      is the order in which they are checked for the template provided
      to the renderer.

    auto_reload
       If true, automatically reload templates when they change (slows
       rendering).  Default: ``True``.

    debug
       If true, show nicer tracebacks during Chameleon template rendering
       errors (slows rendering).  Default: ``True``.

    encoding
       The encoding that the on-disk representation of the templates
       and all non-ASCII values passed to the template should be
       expected to adhere to.  Default: ``utf-8``.

    translator
       A translation function used for internationalization when the
       ``i18n:translate`` attribute syntax is used in the Chameleon
       template is active or a
       :class:`translationstring.TranslationString` is encountered
       during output.  It must accept a translation string and return
       an interpolated translation.  Default: ``None`` (no translation
       performed).
    """
    def __init__(self, search_path, auto_reload=True, debug=False,
                 encoding='utf-8', translator=None):
        self.translate = translator
        loader = ZPTTemplateLoader(search_path=search_path,
                                   auto_reload=auto_reload,
                                   debug=debug,
                                   encoding=encoding,
                                   translate=ChameleonTranslate(translator))
        self.loader = loader

    def __call__(self, template_name, **kw):
        return self.load(template_name)(**kw)

    def load(self, template_name):
        return self.loader.load(template_name + '.pt')


default_dir = resource_filename('nive.components.reform', 'templates/')
default_renderer = ZPTRendererFactory((default_dir,))



# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

from nive.tool import Tool, ToolView
from nive.definitions import ToolConf, FieldConf, ViewConf


configuration = ToolConf(
    id = "exampletool",
    context = "nive.tools.example.tool",
    name = "Empty tool for tests",
    description = "",
    apply = None,  #(IObject,)
    mimetype = "text/html",
)
configuration.data = [
    FieldConf(id="parameter1", datatype="bool",               default=0,  name="Show 1", description="Display 1"),
    FieldConf(id="parameter2", datatype="string", required=1, default="", name="Show 2", description="Display 2")
]
configuration.views = [
    ViewConf(name="", view=ToolView, attr="run", permission="system", context="nive.tools.example.tool")
]


class tool(Tool):

    def _Run(self, **values):
        result = "<h1>OK</h1>"
        return None, result




# -*- coding: utf-8 -*-

import unittest

from nive.utils import country_data
from nive.utils import language


class Country_data(unittest.TestCase):
    
    def test_defs(self):
        for ccc in country_data.countries:
            self.assertTrue(ccc[0], ccc)
            self.assertTrue(ccc[1], ccc)
            self.assertTrue(ccc[2], ccc)
            self.assertTrue(ccc[3], ccc)
                        
        
    def test_codes(self):
        for conf in country_data.GetCountries():
            self.assertTrue(conf.get("id"), conf)
            self.assertTrue(conf.get("name"), conf)
            
            
    def test_get(self):
        self.assertTrue(country_data.GetConf("DEU").get("code2")=="DE")
        self.assertTrue(country_data.GetConf("DE").get("code")=="DEU")
        self.assertTrue(country_data.GetConf("XX").get("code2")==None)
                        
        


class Country(unittest.TestCase):
    

    def test_lang(self):
        cext = language.CountryExtension()
        for c in cext.Codelist():
            self.assertTrue(c.get("id"))
            self.assertTrue(c.get("name"))
            
    def test_conf(self):
        lext = language.CountryExtension()
        for c in country_data.countries:
            conf = lext.GetConf(c[0])
            self.assertTrue(conf.get("code"), c)
            self.assertTrue(conf.get("code2"), c)
            self.assertTrue(conf.get("name"), c)
            

    def test_name(self):
        lext = language.CountryExtension()
        for c in country_data.countries:
            name = lext.GetName(c[0])
            self.assertTrue(name, c)
            
            
            
            
            



from copy import deepcopy
from . import language_data 


class LanguageExtension(object):
    """    
    empty = {}
    empty["code"] = ""
    empty["code2"] = ""
    empty["name"] = ""
    empty["articles_used"] = 0 # 1=article, 2=f+m, 3=f+m+n
    empty["articles"] = {}
    empty["article_abbr"] = ""
    empty["verb_prefix"] = ""
    empty["plural"] = ""
    empty["codepage"] = ""
    empty["remove_chars"] = "\t\r\n"
    empty["special_chars"] = ""
    """
    
    def Codelist(self):
        """
        """
        return language_data.GetLanguages()


    def GetConf(self, lang):
        """
        """
        return language_data.GetConf(lang)


    def GetName(self, lang):
        """
        """
        return language_data.GetConf(lang).get("name")


from . import country_data 

class CountryExtension(object):
    """    
    empty = {}
    empty["code"] = ""
    empty["language"] = ""
    """

    def Codelist(self):
        """
        """
        return country_data.GetCountries()
        

    def GetConf(self, code):
        """
        """
        return country_data.GetConf(code)


    def GetName(self, code):
        """
        """
        return country_data.GetConf(code).get("name")





import time
import unittest

from nive.tools.dbJsonDump import *

from nive.tests import db_app, __local

from nive.helper import FormatConfTestFailure

# -----------------------------------------------------------------

class DBSqlDataTest1(unittest.TestCase):

    def test_conf1(self):
        r=configuration.test()
        if not r:
            return
        self.fail(FormatConfTestFailure(r))


    def test_tool(self):
        dbJsonDump(configuration,None)
        
    
class DBSqlDataTest1_db(__local.DefaultTestCase):

    def setUp(self):
        self._loadApp()
        self.app.Register(configuration)
        root = self.app.root
        o = db_app.createObj1(root)
        db_app.createObj1(o)
        db_app.createObj2(o)
        db_app.createObj3(o)
        self.o = o

    def tearDown(self):
        self._closeApp(True)
    

    def test_toolrun1(self):
        t = self.app.GetTool("nive.tools.dbJsonDump", self.app)
        self.assertTrue(t)
        t.importWf = 0
        t.importSecurity = 0
        r = t()
        #print v
        self.assertTrue(r)


    def test_toolrun2(self):
        t = self.app.GetTool("dbJsonDump", self.app)
        self.assertTrue(t)
        t.importWf = 1
        t.importSecurity = 1
        r = t()
        self.assertTrue(r)




# -*- coding: utf-8 -*-

import unittest

from nive.utils import language_data
from nive.utils import language


class Langugage_data(unittest.TestCase):
    

    def test_defs(self):
        for lang in language_data.languages:
            conf = language_data.GetConf(lang)
            self.assertTrue(conf.get("code"), lang)
            self.assertTrue(conf.get("code2"), lang)
            self.assertTrue(conf.get("name"), lang)
        conf = language_data.GetConf("xxx")
        self.assertTrue(conf.get("code")=="")
        
    def test_codes(self):
        for conf in language_data.GetLanguages():
            self.assertTrue(conf.get("id"), conf)
            self.assertTrue(conf.get("name"), conf)
            
            
    def test_get(self):
        self.assertTrue(language_data.GetConf("ger").get("code2")=="de")
        self.assertTrue(language_data.GetConf("de").get("code")=="ger")
                        
        


class Langugage(unittest.TestCase):
    
    def test_lang(self):
        lext = language.LanguageExtension()
        for c in lext.Codelist():
            self.assertTrue(c.get("id"))
            self.assertTrue(c.get("name"))
            
    def test_conf(self):
        lext = language.LanguageExtension()
        for l in language_data.languages:
            conf = lext.GetConf(l)
            self.assertTrue(conf.get("code"), l)
            self.assertTrue(conf.get("code2"), l)
            self.assertTrue(conf.get("name"), l)
            

    def test_name(self):
        lext = language.LanguageExtension()
        for l in language_data.languages:
            name = lext.GetName(l)
            self.assertTrue(name, l)
            
            
            
            
            


import unittest

def validation_failure_exc(func, *arg, **kw):
    from nive.components.reform.exception import ValidationFailure
    try:
        func(*arg, **kw)
    except ValidationFailure as e:
        return e
    else:
        raise AssertionError('Form error not raised') # pragma: no cover

class TestField(unittest.TestCase):
    def _getTargetClass(self):
        from nive.components.reform.field import Field
        return Field
        
    def _makeOne(self, schema, **kw):
        cls = self._getTargetClass()
        return cls(schema, **kw)

    def test_ctor_defaults(self):
        from nive.components.reform.template import default_renderer
        schema = DummySchema()
        field = self._makeOne(schema)
        self.assertEqual(field.schema, schema)
        self.assertEqual(field.renderer, default_renderer)
        self.assertEqual(field.name, 'name')
        self.assertEqual(field.title, 'title')
        self.assertEqual(field.required, True)
        self.assertEqual(field.order, 0)
        self.assertEqual(field.oid, 'reformField0')
        self.assertEqual(field.children, [])
        self.assertEqual(field.typ, schema.typ)

    def test_ctor_with_children_in_schema(self):
        from nive.components.reform.field import Field
        schema = DummySchema()
        node = DummySchema()
        schema.children = [node]
        field = self._makeOne(schema, renderer='abc')
        self.assertEqual(len(field.children), 1)
        child_field = field.children[0]
        self.assertEqual(child_field.__class__, Field)
        self.assertEqual(child_field.schema, node)
        self.assertEqual(child_field.renderer, 'abc')

    def test_ctor_with_resource_registry(self):
        from nive.components.reform.field import Field
        schema = DummySchema()
        node = DummySchema()
        schema.children = [node]
        field = self._makeOne(schema, resource_registry='abc')
        self.assertEqual(len(field.children), 1)
        child_field = field.children[0]
        self.assertEqual(child_field.__class__, Field)
        self.assertEqual(child_field.schema, node)
        self.assertEqual(child_field.resource_registry, 'abc')

    def test_ctor_with_unknown_kwargs(self):
        from nive.components.reform.field import Field
        schema = DummySchema()
        node = DummySchema()
        schema.children = [node]
        field = self._makeOne(schema, foo='foo', bar='bar')
        self.assertEqual(len(field.children), 1)
        child_field = field.children[0]
        self.assertEqual(field.foo, 'foo')
        self.assertEqual(field.bar, 'bar')
        self.assertEqual(child_field.__class__, Field)
        self.assertEqual(child_field.schema, node)
        self.assertEqual(child_field.foo, 'foo')
        self.assertEqual(child_field.bar, 'bar')

    def test_translate_renderer_has_no_translator(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        self.assertEqual(field.translate('term'), 'term')

    def test_translate_renderer_has_translator(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        field.renderer.translate = lambda foo: 'translated'
        self.assertEqual(field.translate('term'), 'translated')

    def test_set_default_renderer(self):
        cls = self._getTargetClass()
        old = cls.default_renderer
        def new():
            return 'OK'
        try:
            cls.set_default_renderer(new)
            self.assertEqual(cls.default_renderer(), 'OK')
        finally:
            cls.set_default_renderer(old)

    def test_set_default_resource_registry(self):
        cls = self._getTargetClass()
        old = cls.default_resource_registry
        try:
            cls.set_default_resource_registry('OK')
            self.assertEqual(cls.default_resource_registry, 'OK')
        finally:
            cls.set_default_resource_registry(old)

    def test_set_zpt_renderer(self):
        cls = self._getTargetClass()
        old = cls.default_renderer
        from pkg_resources import resource_filename
        template_dir = resource_filename('nive.components.reform', 'templates/')
        class Field:
            oid = None
            name = None
        field = Field()
        try:
            cls.set_zpt_renderer(template_dir)
            self.assertTrue(cls.default_renderer('hidden', field=field,
                                                 cstruct=None))
        finally:
            cls.set_default_renderer(old)

    def test_widget_uses_schema_widget(self):
        widget = DummyWidget()
        schema = DummySchema()
        schema.widget = widget
        schema.typ = DummyType()
        field = self._makeOne(schema)
        widget = field.widget
        self.assertEqual(widget, widget)

    def test_widget_has_maker(self):
        schema = DummySchema()
        def maker():
            return 'a widget'
        schema.typ = DummyType(maker=maker)
        field = self._makeOne(schema)
        widget = field.widget
        self.assertEqual(widget, 'a widget')

    def test_widget_no_maker_no_default_widget_maker(self):
        from nive.components.reform.widget import TextInputWidget
        schema = DummySchema()
        schema.typ = None
        field = self._makeOne(schema)
        widget = field.widget
        self.assertEqual(widget.__class__, TextInputWidget)

    def test_set_widgets_emptystring(self):
        schema = DummySchema()
        field = self._makeOne(schema, renderer='abc')
        widget = DummyWidget()
        field.set_widgets({'':widget})
        self.assertEqual(field.widget, widget)

    def test_set_widgets_emptystring_and_children(self):
        schema = DummySchema()
        field = self._makeOne(schema, renderer='abc')
        child1 = DummyField(name='child1')
        child2 = DummyField(name='child2')
        field.children = [child1, child2]
        widget = DummyWidget()
        widget1 = DummyWidget()
        widget2 = DummyWidget()
        field.set_widgets({'':widget,
                           'child1':widget1,
                           'child2':widget2})
        self.assertEqual(field.widget, widget)
        self.assertEqual(child1.widget, widget1)
        self.assertEqual(child2.widget, widget2)

    def test_set_widgets_childrenonly(self):
        schema = DummySchema()
        field = self._makeOne(schema, renderer='abc')
        child1 = DummyField(name='child1')
        child2 = DummyField(name='child2')
        field.children = [child1, child2]
        widget1 = DummyWidget()
        widget2 = DummyWidget()
        field.set_widgets({'child1':widget1,
                           'child2':widget2})
        self.assertEqual(child1.widget, widget1)
        self.assertEqual(child2.widget, widget2)

    def test_set_widgets_splat(self):
        schema = DummySchema()
        field = self._makeOne(schema, renderer='abc')
        child1 = DummyField(name='child1')
        field.children = [child1]
        widget1 = DummyWidget()
        field.set_widgets({'*':widget1})
        self.assertEqual(child1.widget, widget1)

    def test_set_widgets_nested(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        schema1 = DummySchema()
        schema1.name = 'child1'
        child1 = self._makeOne(schema1)
        schema2 = DummySchema()
        schema2.name = 'child2'
        child2 = self._makeOne(schema2)
        schema3 = DummySchema()
        schema3.name = 'child3'
        child3 = self._makeOne(schema3)
        schema4 = DummySchema()
        schema4.name = 'child4'
        child4 = self._makeOne(schema4)
        field.children = [child1, child2]
        child1.children = [child3]
        child2.children = [child4]
        widget1 = DummyWidget()
        widget2 = DummyWidget()
        widget3 = DummyWidget()
        widget4 = DummyWidget()
        field.set_widgets({'child1':widget1,
                           'child1.child3':widget3,
                           'child2':widget2,
                           'child2.child4':widget4})
        self.assertEqual(child1.widget, widget1)
        self.assertEqual(child2.widget, widget2)
        self.assertEqual(child3.widget, widget3)
        self.assertEqual(child4.widget, widget4)

    def test_set_widgets_complex_nonempty_key_no_children(self):
        schema = DummySchema()
        field = self._makeOne(schema, renderer='abc')
        child1 = DummyField(name='child1')
        child2 = DummyField(name='child2')
        field.children = [child1, child2]
        widget1 = DummyWidget()
        widget2 = DummyWidget()
        field.set_widgets({'child1':widget1,
                           'child2':widget2})
        self.assertEqual(child1.widget, widget1)
        self.assertEqual(child2.widget, widget2)

    def test_get_widget_requirements(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        field.widget.requirements = (('abc', '123'), ('ghi', '789'))
        child1 = DummyField(name='child1')
        field.children = [child1]
        result = field.get_widget_requirements()
        self.assertEqual(result,
                         [('abc', '123'), ('ghi', '789'), ('def', '456')])

    def test_get_widget_resources(self):
        def resource_registry(requirements):
            self.assertEqual(requirements, [ ('abc', '123') ])
            return 'OK'
        schema = DummySchema()
        field = self._makeOne(schema)
        field.widget.requirements = ( ('abc', '123') ,)
        field.resource_registry = resource_registry
        result = field.get_widget_resources()
        self.assertEqual(result, 'OK')

    def test_clone(self):
        schema = DummySchema()
        field = self._makeOne(schema, renderer='abc')
        child = DummyField()
        field.children = [child]
        field.foo = 1
        result = field.clone()
        self.assertFalse(result is field)
        self.assertEqual(result.order, 1)
        self.assertEqual(result.oid, 'reformField1')
        self.assertEqual(result.renderer, 'abc')
        self.assertEqual(result.schema, schema)
        self.assertEqual(result.foo, 1)
        self.assertEqual(result.children, [child])
        self.assertEqual(result.children[0].cloned, True)

    def test___iter__(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        child = DummyField()
        child2 = DummyField()
        field.children = [child, child2]
        result = list(field.__iter__())
        self.assertEqual(result, [child, child2])

    def test___getitem__success(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        child = DummyField()
        field.children = [child]
        self.assertEqual(field['name'], child)
        
    def test___getitem__fail(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        child = DummyField()
        field.children = [child]
        self.assertRaises(KeyError, field.__getitem__, 'nope')

    def test_errormsg_error_None(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        self.assertEqual(field.errormsg, None)
            
    def test_errormsg_error_not_None(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        field.error = DummyInvalid('abc')
        self.assertEqual(field.errormsg, 'abc')

    def test_validate_succeedsItems(self):
        fields = [
            ('name', 'Name'),
            ('title', 'Title'),
            ]
        schema = DummySchema()
        field = self._makeOne(schema)
        field.widget = DummyWidget()
        result = field.validate(fields)
        self.assertEqual(result, {'name':'Name', 'title':'Title'})

    def test_validate_succeeds(self):
        fields = {
            'name': 'Name',
            'title': 'Title',
            }
        schema = DummySchema()
        field = self._makeOne(schema)
        field.widget = DummyWidget()
        result = field.validate(fields)
        self.assertEqual(result, {'name':'Name', 'title':'Title'})

    def test_validate_fails_widgeterror(self):
        from nive.components.reform.schema import Invalid
        fields = {
            'name': 'Name',
            'title': 'Title',
            }
        invalid = Invalid(None, None, dict(fields))
        schema = DummySchema()
        field = self._makeOne(schema)
        field.widget = DummyWidget(exc=invalid)
        e = validation_failure_exc(field.validate, fields)
        self.assertEqual(field.widget.error, invalid)
        self.assertEqual(e.cstruct, dict(fields))
        self.assertEqual(e.field, field)
        self.assertEqual(e.error, invalid)

    def test_validate_fails_schemaerror(self):
        from nive.components.reform.schema import Invalid
        fields = {
            'name': 'Name',
            'title': 'Title',
            }
        invalid = Invalid(None, None)
        schema = DummySchema(invalid)
        field = self._makeOne(schema)
        field.widget = DummyWidget()
        e = validation_failure_exc(field.validate, fields)
        self.assertEqual(field.widget.error, invalid)
        self.assertEqual(e.cstruct, {'name':'Name', 'title':'Title'})
        self.assertEqual(e.field, field)
        self.assertEqual(e.error, invalid)

    def test_validate_fails_widgeterror_and_schemaerror(self):
        from nive.components.reform.schema import Invalid
        fields = {
            'name': 'Name',
            'title': 'Title',
            }
        widget_invalid = Invalid(None, None, dict(fields))
        schema_invalid = Invalid(None, None)
        schema = DummySchema(schema_invalid)
        field = self._makeOne(schema)
        field.widget = DummyWidget(exc=widget_invalid)
        e = validation_failure_exc(field.validate, fields)
        self.assertEqual(field.widget.error.msg, schema_invalid.msg)
        self.assertEqual(e.cstruct, dict(fields))
        self.assertEqual(e.field, field)
        self.assertEqual(e.error.msg, schema_invalid.msg)

    def test_render(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        widget = field.widget = DummyWidget()
        self.assertEqual(field.render('abc'), 'abc')
        self.assertEqual(widget.rendered, 'writable')

    def test_serialize(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        widget = field.widget = DummyWidget()
        self.assertEqual(field.serialize('abc'), 'abc')
        self.assertEqual(widget.rendered, 'writable')

    def test_serialize_null(self):
        from nive.components.reform.schema import null
        schema = DummySchema()
        field = self._makeOne(schema)
        widget = field.widget = DummyWidget()
        self.assertEqual(field.serialize(null), null)
        self.assertEqual(widget.rendered, 'writable')

    def test_deserialize(self):
        cstruct = {'name':'Name', 'title':'Title'}
        schema = DummySchema()
        field = self._makeOne(schema)
        field.widget = DummyWidget()
        result = field.deserialize(cstruct)
        self.assertEqual(result, {'name':'Name', 'title':'Title'})

    def test___repr__(self):
        schema = DummySchema()
        field = self._makeOne(schema)
        r = repr(field)
        self.assertTrue(r.startswith('<nive.components.reform.field.Field object at '), r)
        self.assertTrue(r.endswith("(schemanode 'name')>"))

    def test_configuration(self):
        from nive.definitions import FieldConf
        schema = DummySchema()
        field = self._makeOne(schema, configuration=FieldConf())
        self.assertTrue(field.configuration)
        field = self._makeOne(schema)
        self.assertFalse(field.configuration)

class DummyField(object):
    oid = 'oid'
    requirements = ( ('abc', '123'), ('def', '456'))
    required = False
    def __init__(self, schema=None, renderer=None, name='name'):
        self.schema = schema
        self.renderer = renderer
        self.name = name

    def clone(self):
        self.cloned = True
        return self

    def get_widget_requirements(self, L=None):
        return self.requirements

class DummySchema(object):
    typ = None
    name = 'name'
    title = 'title'
    description = 'description'
    required = True
    children = ()
    default = 'default'
    sdefault = 'sdefault'
    def __init__(self, exc=None):
        self.exc = exc
    def deserialize(self, value, formstruct=None):
        if self.exc:
            raise self.exc
        return value

    def serialize(self, value):
        return value

class DummyType(object):
    def __init__(self, maker=None):
        self.widget_maker = maker
        
class DummyWidget(object):
    rendered = None
    def __init__(self, exc=None):
        self.exc = exc

    def deserialize(self, field, pstruct, formstruct=None):
        if self.exc is not None:
            raise self.exc
        return pstruct

    def serialize(self, field, cstruct=None):
        self.rendered = 'writable'
        return cstruct

    def handle_error(self, field, e):
        self.error = e

class DummyInvalid(object):
    def __init__(self, msg=None):
        self.msg = msg



# -*- coding: utf-8 -*-

import unittest
import os
import tempfile

from nive.utils.path import DvPath



class Path(unittest.TestCase):
    
    def setUp(self):
        self.base = os.sep+"tmp_nivepathtest_000"+os.sep
        self.name = self.base+"nofile.txt"
        pass
    
    def test_Create(self):
        n = self.name
        p = DvPath(n)
        self.assertTrue(p._path==n)
        self.assertTrue(str(p)==n)
        p = DvPath(DvPath(n))
        self.assertTrue(str(p)==n)
        p.SetStr(n)
        self.assertTrue(str(p)==n)
        self.assertTrue(p.GetStr()==n)
        
        
    def test_Set(self):
        n = self.name
        p = DvPath(n)
        self.assertTrue(str(p)==n)
        p.SetName("new")
        self.assertTrue(str(p)==self.base+"new.txt", str(p))
        p.SetExtension("png")
        self.assertTrue(str(p)==self.base+"new.png")
        p.SetNameExtension("another.txt")
        self.assertTrue(str(p)==self.base+"another.txt")
        

    def test_Dirs(self):
        n = self.base
        p = DvPath(n)
        self.assertTrue(str(p)==n)
        p.Append("another_dir"+os.sep+"and.file")
        self.assertTrue(str(p)==n+"another_dir"+os.sep+"and.file", str(p))
        p = DvPath(n)
        p.AppendDirectory("the_last")
        self.assertTrue(str(p)==n+"the_last"+os.sep, str(p))
        p.RemoveLastDirectory()
        self.assertTrue(str(p)==n)
        p = DvPath(n[:-1])
        p.AppendSeperator()
        self.assertTrue(str(p)==self.base)
        
        
            
    def test_Get(self):
        n = self.name
        p = DvPath(n)
        self.assertTrue(p.GetPath() == self.base)
        self.assertTrue(p.GetName() == "nofile")
        self.assertTrue(p.GetExtension() == "txt")
        self.assertTrue(p.GetNameExtension() == "nofile.txt")
        self.assertTrue(p.GetSize() == -1)
        self.assertTrue(p.GetLastDirectory() == "tmp_nivepathtest_000")
        p.IsFile()
        p.IsDirectory()
        p.Exists()
        p.IsValid()

    
    def test_fncs(self):
        temp = tempfile.gettempdir()
        p = DvPath(temp)
        p.AppendDirectory("tmp_nivepathtest_000")
        p.Delete(deleteSubdirs = True)
        p.CreateDirectories()
        self.assertTrue(p.IsDirectory())
        p.CreateDirectoriesExcp()
        p.Rename("tmp_nivepathtest_111")
        p.AppendDirectory("tmp_nivepathtest_000")
        p.AppendDirectory("tmp_nivepathtest_000")
        p.CreateDirectories()
        self.assertTrue(p.IsDirectory())
        p = DvPath(temp)
        p.AppendDirectory("tmp_nivepathtest_000")
        p.Delete(deleteSubdirs = False)
        self.assertTrue(p.IsDirectory()==True)
        p.Delete(deleteSubdirs = True)
        self.assertTrue(p.IsDirectory()==False)
        #p.Execute("cmd", returnStream = True)
        
        
    def test_tempfilename(self):
        p = DvPath("file.txt")
        p.SetUniqueTempFileName()
        self.assertTrue(p.GetExtension()=="txt",p.GetExtension())
        
            
            


# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

__doc__ = """
Data Pool MySql Module
----------------------
*Requires MySQL-python*. Use 'pip install MySQL-python' to install the package.
"""


import threading
from nive.definitions import OperationalError, ConfigurationError

try:
    import MySQLdb
except ImportError:
    # make tests pass without mysql package
    class MySQLdb(object):
        """ """
        OperationalError = None
        ProgrammingError = None
        Warning = None
        @staticmethod
        def connect(*args,**kw):
            raise ImportError("Python MySQLdb binding not available. Try 'pip install MySQL-python' to install the package.")
    
from nive.utils.utils import STACKF

from nive.utils.dataPool2.base import Base, Entry
from nive.utils.dataPool2.base import NotFound
from nive.utils.dataPool2.connection import Connection, ConnectionThreadLocal, ConnectionRequest

from nive.utils.dataPool2.mysql.dbManager import MySQLManager
from nive.utils.dataPool2.files import FileManager, FileEntry



class MySqlConnection(Connection):
    """
    MySql connection handling class

    config parameter in dictionary:
    user = database user
    password = password user
    host = host mysql server
    port = port mysql server
    db = initial database name

    timeout is set to 3.
    """
    # todo [3] unicode ?

    def connect(self):
        """ Close and connect to server """
        #t = time()
        conf = self.configuration
        use_unicode = conf.unicode
        charset = None
        if use_unicode:
            charset = "utf8"
        db = MySQLdb.connect(conf.host,
                             conf.user,
                             str(conf.password), 
                             port=conf.port or 3306,
                             db=conf.dbName, 
                             connect_timeout=conf.timeout, 
                             use_unicode=use_unicode, 
                             charset=charset)
        if not db:
            raise OperationalError("Cannot connect to database '%s.%s'" % (conf.host, conf.dbName))
        self._set(db)
        #print "connect:", time() - t
        return db
        

    def IsConnected(self):
        """ Check if database is connected """
        try:
            db = self._get()
            db.ping()
            return db.cursor()!=None
        except:
            return False


    def GetDBManager(self):
        """ returns the database manager obj """
        self.VerifyConnection()
        aDB = MySQLManager()
        aDB.SetDB(self.PrivateConnection())
        return aDB


    def FmtParam(self, param):
        """format a parameter for sql queries like literal for mysql db"""
        db = self._get()
        if not db:
            self.connect()
            db = self._get()
        return db.literal(param)


    def PrivateConnection(self):
        """ This function will return a new and raw connection. It is up to the caller to close this connection. """
        conf = self.configuration
        use_unicode = conf.unicode
        charset = None
        if use_unicode:
            charset = "utf8"
        db = MySQLdb.connect(conf.host,
                             conf.user,
                             str(conf.password), 
                             port=conf.port or 3306,
                             db=conf.dbName, 
                             connect_timeout=conf.timeout, 
                             use_unicode=use_unicode, 
                             charset=charset)
        return db



class MySqlConnThreadLocal(MySqlConnection, ConnectionThreadLocal):
    """
    Caches database connections as thread local values.
    """

    def __init__(self, config = None, connectNow = True):
        self.local = threading.local()
        MySqlConnection.__init__(self, config, connectNow)


class MySqlConnRequest(MySqlConnection, ConnectionRequest):
    """
    Caches database connections as request values. Uses thread local stack as fallback (e.g testing).
    """

    def __init__(self, config = None, connectNow = True):
        self.local = threading.local()
        MySqlConnection.__init__(self, config, connectNow)



class MySql(FileManager, Base):
    """
    Data Pool MySql 5 implementation
    """
    _OperationalError = MySQLdb.OperationalError
    _ProgrammingError = MySQLdb.ProgrammingError
    _Warning = MySQLdb.Warning
    _DefaultConnection = MySqlConnRequest


    def _GetInsertIDValue(self, cursor):
        cursor.execute("SELECT LAST_INSERT_ID()")
        return cursor.fetchone()[0]

               
    def _CreateNewID(self, table = "", dataTbl = None):
        #
        aC = self.connection.cursor()
        if table == "":
            table = self.MetaTable
        if table == self.MetaTable:
            if not dataTbl:
                raise ConfigurationError("Missing data table - Entry not created")

            # sql insert empty rec in meta table
            if self._debug:
                STACKF(0,"INSERT INTO %s () VALUES ()" % (dataTbl)+"\r\n",self._debug, self._log,name=self.name)
            try:
                aC.execute("INSERT INTO %s () VALUES ()" % (dataTbl))
            except self._Warning:
                pass
            # sql get id of created rec
            if self._debug:
                STACKF(0,"SELECT LAST_INSERT_ID()\r\n",0, self._log,name=self.name)
            aC.execute("SELECT LAST_INSERT_ID()")
            dataref = aC.fetchone()[0]
            # sql insert empty rec in meta table
            if self._debug:
                STACKF(0,"INSERT INTO %s (pool_datatbl, pool_dataref) VALUES ('%s', %s)"% (self.MetaTable, dataTbl, dataref),0, self._log,name=self.name)
            aC.execute("INSERT INTO %s (pool_datatbl, pool_dataref) VALUES ('%s', %s)"% (self.MetaTable, dataTbl, dataref))
            if self._debug:
                STACKF(0,"SELECT LAST_INSERT_ID()\r\n",0, self._log,name=self.name)
            aC.execute("SELECT LAST_INSERT_ID()")
            aID = aC.fetchone()[0]
            aC.close()
            return aID, dataref

        # sql insert empty rec in meta table
        if self._debug:
            STACKF(0,"INSERT INTO %s () VALUES ()" % (table)+"\r\n",self._debug, self._log,name=self.name)
        aC.execute("INSERT INTO %s () VALUES ()" % (table))
        # sql get id of created rec
        if self._debug:
            STACKF(0,"SELECT LAST_INSERT_ID()\r\n",0, self._log,name=self.name)
        aC.execute("SELECT LAST_INSERT_ID()")
        aID = aC.fetchone()[0]
        aC.close()
        return aID, 0


    # types/classes -------------------------------------------------------------------

    def _GetPoolEntry(self, id, **kw):
        try:
            return MySqlEntry(self, id, **kw)
        except NotFound:
            return None



    # MySql 4 tree structure --------------------------------------------------------------

    def GetParentPath4(self, id):
        """
        MySql 4 version
        returns id references of parents for id.
        """
        sql = """
        SELECT p0.pool_unitref as ref1,p1.pool_unitref as ref2,p2.pool_unitref as ref3,p3.pool_unitref as ref4,p4.pool_unitref as ref5,p5.pool_unitref as ref6,p6.pool_unitref as ref7
        FROM %(meta)s p0, %(meta)s p1, %(meta)s p2, %(meta)s p3, %(meta)s p4, %(meta)s p5, %(meta)s p6
        WHERE  p0.id = %(id)d
        and    p1.id = p0.pool_unitref
        and IF(p1.pool_unitref > 0, p2.id = p1.pool_unitref, p2.id = p0.pool_unitref)
        and IF(p2.pool_unitref > 0, p3.id = p2.pool_unitref, p3.id = p0.pool_unitref)
        and IF(p3.pool_unitref > 0, p4.id = p3.pool_unitref, p4.id = p0.pool_unitref)
        and IF(p4.pool_unitref > 0, p5.id = p4.pool_unitref, p5.id = p0.pool_unitref)
        and IF(p5.pool_unitref > 0, p6.id = p5.pool_unitref, p6.id = p0.pool_unitref)
        Group by ref1""" % {"meta":self.MetaTable, "id":int(id)}
        t = self.Query(sql)
        parents = []
        if len(t) == 0:
            return parents
        cnt = 7
        for i in range(0, cnt-1):
            id = t[0]
            if id == 0:
                break
            parents.insert(0, id)
        return parents


    def GetParentTitles4(self, id):
        """
        MySql 4 version
        returns titles of parents for id.
        """
        sql = """
        SELECT p0.pool_unitref as ref1,p0.title as t1,p1.pool_unitref as ref2,p1.title as t2,p2.pool_unitref as ref3,p2.title as t3,p3.pool_unitref as ref4,p3.title as t4,p4.pool_unitref as ref5,p4.title as t5,p5.pool_unitref as ref6,p5.title as t6,p6.pool_unitref as ref7,p6.title as t7
        FROM %(meta)s p0, %(meta)s p1, %(meta)s p2, %(meta)s p3, %(meta)s p4, %(meta)s p5, %(meta)s p6
        WHERE  p0.id = %(id)d
        and    p1.id = p0.pool_unitref
        and IF(p1.pool_unitref > 0, p2.id = p1.pool_unitref, p2.id = p0.pool_unitref)
        and IF(p2.pool_unitref > 0, p3.id = p2.pool_unitref, p3.id = p0.pool_unitref)
        and IF(p3.pool_unitref > 0, p4.id = p3.pool_unitref, p4.id = p0.pool_unitref)
        and IF(p4.pool_unitref > 0, p5.id = p4.pool_unitref, p5.id = p0.pool_unitref)
        and IF(p5.pool_unitref > 0, p6.id = p5.pool_unitref, p6.id = p0.pool_unitref)
        Group by ref1""" % {"meta":self.MetaTable, "id":int(id)}
        t = self.Query(sql)
        parents = []
        if len(t) == 0:
            return parents
        cnt = 4
        for i in range(1, cnt-1):
            title = t[0][i*2+1]
            parents.insert(0, self.EncodeText(title))
            if t[0][i*2] == 0:
                break
        return parents


    def _FmtListForQuery(self, values):
        return values
        #FmtParam = self.connection.FmtParam
        #return ",".join([FmtParam(v) for v in values])


class MySqlEntry(FileEntry, Entry):
    """
    Data Pool Entry MySql 5 implementation
    """






import time
import unittest
import logging
from smtplib import SMTPServerDisconnected

from nive.definitions import *
from nive.tools.sendMail import sendMail, configuration

from nive.tests import __local


from nive.helper import FormatConfTestFailure



# -----------------------------------------------------------------

class SendMailTest1(unittest.TestCase):


    def test_conf1(self):
        r=configuration.test()
        if not r:
            return
        self.fail(FormatConfTestFailure(r))


    def test_tool(self):
        sendMail(configuration,None)
        
        
class SendMailTest2_db(__local.DefaultTestCase):

    def setUp(self):
        self._loadApp()
        self.app.Register(configuration)
        logging.basicConfig()
    
    def tearDown(self):
        self._closeApp(True)


    def test_tool(self):
        t = self.app.GetTool("sendMail")
        self.assertTrue(t)

    
    def test_toolrun1(self):
        t = self.app.GetTool("sendMail")
        self.assertTrue(t)
        try:
            r = t()
        except ConfigurationError:
            pass


    def test_toolrun2(self):
        t = self.app.GetTool("sendMail")
        self.assertTrue(t)
        try:
            r = t(recvmails=[("test@aaaaaaaa.com", "No name")], title="Testmail", body="body mail")
        except ConfigurationError:
            pass


    def test_toolrundebug(self):
        t = self.app.GetTool("sendMail")
        self.assertTrue(t)
        try:
            r = t(debug=1, recvmails=[("test@aaaaaaaa.com", "No name")], title="Testmail", body="body mail")
        except ConfigurationError:
            pass


    def test_date(self):
        t = self.app.GetTool("sendMail")
        date = t._FormatDate()
        self.assertTrue(date)


    def test_mailstr(self):
        t = self.app.GetTool("sendMail")
        self.assertTrue(t._GetMailStr("aaa@ddd.aa")=="aaa@ddd.aa")
        self.assertEqual(t._GetMailStr(("aaa@ddd.aa","ä a")), '"=?utf-8?q?=C3=A4_a?=" <aaa@ddd.aa>')
        self.assertTrue(t._GetMailStr(("aaa@ddd.aa",))=="aaa@ddd.aa")


    def test_message(self):
        t = self.app.GetTool("sendMail")
        values = dict(contentType="text/html;charset=utf-8",
                      body="aaaaa",
                      fromName="uuuu",
                      fromMail="uuu@aaa.dd",
                      to=["ooo", "ggg"],
                      cc=["ooo", "ggg"],
                      bcc=["ooo", "ggg"],
                      sender="mmmm",
                      replyTo="uoah",
                      subject="My mail")
        m = t._PrepareMessage(**values)
        self.assertTrue(str(m))

        values = dict(contentType="text/html;charset=utf-8",
                      body="aaaaa",
                      fromMail="uuu@aaa.dd",
                      to="ooo",
                      cc="ooo",
                      bcc="ooo",
                      subject="My mail")
        m = t._PrepareMessage(**values)
        self.assertTrue(str(m))



# Copyright 2012, 2013 Arndt Droullier, Nive GmbH. All rights reserved.
# Released under GPL3. See license.txt
#

__doc__ = "Data Pool 2 SQL Base Module"

import weakref
import logging

from datetime import datetime, date

from nive.utils.utils import ConvertToDateTime
from nive.utils.utils import ConvertListToStr
from nive.utils.utils import STACKF
from nive.utils.path import DvPath

from nive.definitions import ConfigurationError, OperationalError, ProgrammingError, ConnectionError, Warning

from nive.utils.dataPool2.structure import PoolStructure, DataWrapper, MetaWrapper, FileWrapper



class Base(object):
    """
    Data Pool 2 SQL Base implementation

    Manage typed data units consisting of meta layer, data layer and files.
    Meta layer is the same for all units, data layer is based on types and files
    are stored by key.
    Meta and data is mapped to database, files are stored in filesystem.

    Pool structure is required and must contain the used database fields as list
    (table names and column names).

    Entries are referenced by pool_meta.id as unique ID.

    All basic SQL is handled internally.

    Configuration Parameter
    -----------------------
    root:          string. filesystem root path
    codePage:      string. the codepage used on output
    dbCodePage:    string. database codepage
    structure:     PoolStructure object defining the database structure and field types.
    version:       string. the default version
    useBackups:    bool.     store backup versions of files on replace
    useTrashcan:   bool.     moves files to trashcan rather than delete physically
    debug:         number. turn debugging on. 0 = off, 1=on (no traceback), 2...20=on (traceback lines) 
    log:           string. log file path for debugging

    Standard Meta
    -------------
    A list consisting of meta layer fields which are used in preloading

    "meta_flds": (
    ("pool_dataref", "integer not null default 0"),
    ("pool_datatbl", "ENUM() not null"),
    )
    """
    # map db system exceptions to one type
    _OperationalError=OperationalError
    _ProgrammingError=ProgrammingError
    _Warning=Warning
    _DefaultConnection = None
    _EmptyValues = None

    MetaTable = "pool_meta"
    FulltextTable = "pool_fulltext"
    GroupsTable = "pool_groups"
    

    def __init__(self, connection = None, structure = None, root = "",
                 useTrashcan = False, useBackups = False, 
                 codePage = "utf-8", dbCodePage = "utf-8",
                 connParam = None,
                 debug = 0, log = "sql.log",
                 timezone = None, **kw):

        self.codePage = codePage
        self.dbCodePage = dbCodePage
        self.useBackups = useBackups
        self.useTrashcan = useTrashcan
        self.pytimezone = timezone

        self._debug = debug
        self._log = log
        self.name = root        # used for logging
        
        self._conn = None
        if not structure:
            self.structure = self._GetDefaultPoolStructure()(pool=self)
        else:
            self.structure = structure
        
        # create directories, if not available
        self.InitFileStorage(root, connParam)
        if connection:
            self._conn = connection
            self.name = self._conn.configuration.dbName
        elif connParam:
            self._conn = self.CreateConnection(connParam)
            self.name = connParam.get("dbName","")
        

    def Close(self):
        if self._conn is not None:
            self._conn.close()

    def __del__(self):
         self.Close()


    # Database api ---------------------------------------------------------------------

    @property
    def connection(self):
        """
        Returns the database connection and attempts a reconnect or verifies the connection
        depending on configuration settings.
        """
        self._conn.VerifyConnection()
        return self._conn

    @property
    def usedconnection(self):
        """
        Returns the previously used connection without verifying the connection. The returned
        connection may be none or not connected. Can be used in cases if the connection will
        only be used to finish previous actions e.g. Commit(). Lookup is faster and will not 
        waste unnecessary resources.
        """
        return self._conn

    @property
    def dbapi(self):
        if not self._conn:
            raise ConnectionError("No Connection")
        return self._conn.dbapi


    def Begin(self):
        """
        Start a database transaction, if supported
        """
        self.connection.begin()

    def Undo(self):
        """
        Rollback the changes made to the database, if supported
        """
        if self.usedconnection:
            self.usedconnection.rollback()

    def Commit(self, user=""):
        """
        Commit the changes made to the database, if supported
        """
        self.usedconnection.commit()

    def GetDBDate(self, date=None):
        if not date:
            date = datetime.now(tz=self.pytimezone)
        elif not isinstance(date, datetime):
            date = ConvertToDateTime(date)
        return date.strftime("%Y-%m-%d %H:%M:%S")
    
    @property
    def placeholder(self):
        """
        SQL statement format string to be used as placeholder for values. 
        e.g. mysql=%s, sqlite=?
        """
        return self._conn.placeholder
    

    # SQL queries ---------------------------------------------------------------------

    def FmtSQLSelect(self, flds, parameter=None, dataTable="", start=0, max=0, **kw):
        """
        Create a select statement based on pool_structure and given parameters.
        
        Aggregate functions can be included in flds but must be marked with
        the prefix `-` e.g. `-count(*)`
        
        Tables in the generated statement get an alias prefix:
        meta__ for meta table
        data__ for data table
        
        flds: select fields
        parameter: where clause parameter
        dataTable: add join statement for the table
        start: start number of record to be returned
        max: maximum nubers of records in result

        **kw:
        singleTable: 1/0 skip join. only use datatable for select
        version: version key. select content from specific version
        operators: operators used for fields in where clause: =,LIKE,<,>,...
        jointype: type of join. default = INNER
        logicalOperator = link between conditions. default = and. and, or, not
        condition = custom condition statement. default = empty
        join = custom join statement. default = empty
        groupby: add GROUP BY statement
        sort: result sort order. sort prefix ! to skip auto table lookup
        ascending: result sort order ascending or descending
        """
        operators = kw.get("operators",{})
        jointype = operators.get("jointype", "INNER")
        singleTable = kw.get("singleTable",0)
        version = kw.get("version")
        metaStructure = self.structure.get(self.MetaTable, version=version)
        mapJoinFld = kw.get("mapJoinFld")
        plist = []   # sorted list of query parameters for execute()
        ph = self.placeholder   # placeholder to be used instead plist values
        fields = []
        for field in flds:
            if singleTable:
                table = ""
                if field[0] == "-":
                    field = field[1:]
            else:
                table = "meta__."
                # aggregate functions marked with - e.g. "-count(*)"
                if field[0] == "-":
                    table = ""
                    field = field[1:]
                asName = field.find(" as ")!=-1
                f2=field
                if asName:
                    f2 = field.split(" as ")[0]
                elif f2 in metaStructure:   # ?? or f2 == "pool_stag":
                    table = "meta__."
                elif table != "":
                    if jointype.lower() != "inner":
                        if field == mapJoinFld:
                            fields.append("IF(meta__.pool_datatbl='%s', %s, meta__.title)" % (dataTable, field))
                        else:
                            fields.append("IF(meta__.pool_datatbl='%s', %s, '')" % (dataTable, field))
                        continue
                    table = "data__."
            #if len(fields) > 0:
            #    fields.append(", ")
            fields.append(table + field)
        fields = ",".join(fields)

        aCombi = kw.get("logicalOperator")
        if not aCombi:
            aCombi = "AND"
        else:
            aCombi = aCombi.upper()
            if not aCombi in ("AND","OR","NOT"):
                aCombi = "AND"
        addCombi = False
        
        where = []
        if parameter is None:
            parameter={}
        for key in list(parameter.keys()):
            value = parameter[key]
            paramname = key

            operator = "="
            if operators and key in operators:
                operator = operators[key]

            if singleTable:
                table = ""
            else:
                table = "meta__."
                # aggregate functions marked with - e.g. "-count(*)"
                if key[0] == "-":
                    table = ""
                    paramname = key[1:]
                elif key in metaStructure:
                    table = "meta__."
                elif dataTable != "":
                    table = "data__."
            
            # fmt string values
            if isinstance(value, str):
                if operator == "LIKE":
                    if value == "":
                        continue
                    value = "%%%s%%" % value.replace("*", "%")
                    if addCombi:
                        where.append(" %s " % aCombi)
                    where.append("%s%s %s %s " % (table, paramname, operator, ph))
                    plist.append(value)
                elif operator == "BETWEEN":
                    if value == "":
                        continue
                    if addCombi:
                        where.append(" %s " % aCombi)
                    where.append("%s%s %s %s " % (table, paramname, operator, ph))
                    plist.append(value)
                elif operator == "IN":
                    operator = "="
                    if addCombi:
                        where.append(" %s " % aCombi)
                    where.append("%s%s %s %s " % (table, paramname, operator, ph))
                    plist.append(value)
                else:
                    if addCombi:
                        where.append(" %s " % aCombi)
                    where.append("%s%s %s %s " % (table, paramname, operator, ph))
                    plist.append(value)
                addCombi = True

            # fmt list values
            elif isinstance(value, (tuple, list)):
                if not value:
                    continue
                if addCombi:
                    where.append(" %s " % aCombi)
                if operator == "BETWEEN":
                    if len(value) < 2:
                        continue
                    where.append("%s%s %s %s AND %s" % (table, paramname, operator, ph, ph))
                    plist.append(value[0])
                    plist.append(value[1])
                elif operator.startswith("LIKE:"):
                    if value == "":
                        continue
                    if operator.endswith("OR"):
                        logicalOp = "OR"
                    elif operator.endswith("NOT"):
                        logicalOp = "NOT"
                    else:
                        logicalOp = "AND"
                    where.append(" (")
                    statement = False
                    for v in value:
                        # handle multiple like values
                        if statement:
                            where.append(" %s " % logicalOp)
                        statement=True
                        # exception empty value
                        if not v:
                            where.append("%s%s %s %s " % (table, paramname, "=", ph))
                        else:
                            v = "%%%s%%" % v.replace("*", "%")
                            where.append("%s%s %s %s " % (table, paramname, "LIKE", ph))
                        plist.append(v)
                    where.append(") ")
                elif len(value)==1:
                    if operator == "IN":
                        operator = "="
                    elif operator == "NOT IN":
                        operator = "<>"
                    where.append("%s%s %s %s " % (table, paramname, operator, ph))
                    plist.append(value[0])
                else:
                    v = self._FmtListForQuery(value)
                    if isinstance(v, str):
                        # sqlite error: cannot use placeholder with lists
                        where.append("%s%s %s (%s) " % (table, paramname, operator, v))
                    else:
                        where.append("%s%s %s %s " % (table, paramname, operator, ph))
                        plist.append(value)
                addCombi = True

            # fmt number values
            elif isinstance(value, (int, float)):
                if addCombi:
                    where.append(" %s " % aCombi)
                if operator == "LIKE":
                    operator = "="
                where.append("%s%s %s %s" % (table, paramname, operator, ph))
                plist.append(value)
                addCombi = True

            # fmt datetime values
            elif isinstance(value, datetime):
                if addCombi:
                    where.append(" %s " % aCombi)
                if operator == "LIKE":
                    operator = "="
                where.append("%s%s %s %s" % (table, paramname, operator, ph))
                plist.append(value)
                addCombi = True

            # fmt datetime values
            elif isinstance(value, date):
                if addCombi:
                    where.append(" %s " % aCombi)
                if operator == "LIKE":
                    operator = "="
                where.append("DATE(%s%s) %s %s" % (table, paramname, operator, ph))
                plist.append(value)
                addCombi = True

        condition = kw.get("condition")
        if condition != None and condition != "":
            if len(where):
                where.append(" %s %s" %(aCombi, condition))
            else:
                where = [condition]
        where = self._FmtWhereClause(where, singleTable)

        order = "ASC"
        if kw.get("ascending", 1) == 0:
            order = "DESC"

        # map sort fields
        sort = kw.get("sort", "")
        if sort is None:
            sort = ""
        if sort and sort[0] == "!":
            sort = sort[1:]
        elif sort and not singleTable:
            table = "meta__."
            s = sort.split(",")
            sort = []
            for sortfld in s:
                sortfld = sortfld.strip(" ")
                sortfldvalue = sortfld.split(" ")[0]
                if len(sortfld) > 0 and sortfld[0] == "-":
                    table = ""
                    sortfld = sortfld[1:]
                elif sortfldvalue in metaStructure or sortfldvalue == "pool_stag":
                    table = "meta__."
                elif dataTable != "":
                    table = "data__."
                if len(sort):
                    sort.append(", ")
                sort.append(table)
                sort.append(sortfld)
            sort = "".join(sort)

        customJoin = kw.get("join", "")
        if customJoin is None:
            customJoin = ""

        join = ""
        joindata = ""
        if not singleTable:
            # joins
            if dataTable != "":
                joindata = "%s JOIN %s AS data__ ON (meta__.pool_dataref = data__.id) " % (jointype, dataTable)

        limit = ""
        if max:
            limit = self._FmtLimit(start, max)

        groupby = ""
        if kw.get("groupby"):
            groupby = "GROUP BY %s" % kw.get("groupby")

        table = self.MetaTable + " AS meta__"
        if singleTable:
            table = dataTable

        if sort != "":
            sort = "ORDER BY %s %s" % (sort, order)

        sql = """
        SELECT %s
        FROM %s
        %s
        %s
        %s
        %s
        %s
        %s
        %s
        """ % (fields, table, join, joindata, customJoin, where, groupby, sort, limit)
        return sql, plist


    def _FmtListForQuery(self, values):
        return values

    def _FmtWhereClause(self, where, singleTable):
        if len(where):
            where = "WHERE %s" % "".join(where)
        else:
            where = ""
        return where

    def _FmtLimit(self, start, max):
        if start != None:
            return "LIMIT %s, %s" % (str(start), str(max))
        return "LIMIT %s" % (str(max))


    def GetFulltextSQL(self, searchPhrase, flds, parameter, dataTable = "", **kw):
        """
        generate sql statement for fulltext query

        searchPhrase: text to be searched

        For further options see -> FmtSQLSelect
        """
        sql, values = self.FmtSQLSelect(flds, parameter, dataTable=dataTable, **kw)
        searchPhrase = self.DecodeText(searchPhrase)
        values.insert(0, searchPhrase)
        ph = self.placeholder
        phrase = """%s.text LIKE %s""" % (self.FulltextTable, ph)

        if not searchPhrase in ("", "'%%'", None):
            if sql.find("WHERE ") == -1:
                if sql.find("ORDER BY ") != -1:
                    sql = sql.replace("ORDER BY ", "WHERE %s \r\n        ORDER BY " % (phrase))
                elif sql.find("LIMIT ") != -1:
                    sql = sql.replace("LIMIT ", "WHERE %s \r\n        LIMIT " % (phrase))
                else:
                    sql += "WHERE %s " % (phrase)
            else:
                sql = sql.replace("WHERE ", "WHERE %s AND " % (phrase))

        sql = sql.replace("FROM %s AS meta__"%(self.MetaTable), "FROM %s AS meta__\r\n        LEFT JOIN %s ON (meta__.id = %s.id)"%(self.MetaTable, self.FulltextTable, self.FulltextTable))
        return sql, values
    
    
    # database query execution -----------------------------------------------------------------------

    def Execute(self, sql, values = None, cursor = None):
        """
        Execute a query on the database. Returns the dbapi cursor. Use `cursor.fetchall()` or
        `cursor.fetchone()` to retrieve results. The cursor should be closed after usage.
        """
        if cursor is None:
            cursor = self.connection.cursor()
        if self._debug:
            STACKF(0,sql+"\r\n",self._debug, self._log,name=self.name)
        # adjust different accepted empty values sets
        if not values:
            values = self._EmptyValues
        try:
            cursor.execute(sql, values)
        except self._OperationalError as e:
            # map to nive.utils.dataPool2.base.OperationalError
            raise OperationalError(e)
        except self._ProgrammingError as e:
            # map to nive.utils.dataPool2.base.OperationalError
            self.Undo()
            raise ProgrammingError(e)
        return cursor

    
    def Query(self, sql, values = None, getResult = True):
        """
        execute a query on the database. non unicode texts are converted according to codepage settings.
        """
        c = self.connection.cursor()
        if self._debug:
            STACKF(0,sql+"\r\n",self._debug, self._log,name=self.name)
        sql = self.DecodeText(sql)
        if values:
            # check for list and strings
            found = [isinstance(v, list) or isinstance(v, bytes) for v in values]
            if True in found:
                # shouldnt happen
                v1 = []
                for v in values:
                    if isinstance(v, list):
                        v1.append(tuple(v))
                    elif isinstance(v, bytes):
                        v1.append(self.DecodeText(v))
                    else:
                        v1.append(v)
                values = v1
        # adjust different accepted empty values sets
        if not values:
            values = self._EmptyValues
        try:
            c.execute(sql, values)
        except self._OperationalError as e:
            # map to nive.utils.dataPool2.base.OperationalError
            self.Undo()
            logging.getLogger(self.name).error(str(e) + "  " + sql)
            raise OperationalError(e)
        except self._ProgrammingError as e:
            # map to nive.utils.dataPool2.base.OperationalError
            self.Undo()
            logging.getLogger(self.name).error(str(e) + "  " + sql)
            raise ProgrammingError(e)
        if not getResult:
            c.close()
            return
        result = c.fetchall()
        c.close()
        return result


    def SelectFields(self, table, fields, idValues, idColumn = None):
        """
        Select row with multiple fields in the table.
        Set `idColumn` to the column name of the unique id column
        
        table: table name
        fields: list of field names to be returned
        idValues: list of matching id values to be returned
        idColumn: default 'id'. id column name

        returns matching records 
        """
        dataList = []
        phdata = []
        ph = self.placeholder
        key = idColumn or "id"
        for value in idValues:
            phdata.append(ph)
            dataList.append(value)

        sql = "SELECT %s FROM %s WHERE %s IN (%s)" % (",".join(fields), table, key, ",".join(phdata))

        if self._debug:
            STACKF(0,sql+"\r\n",self._debug, self._log,name=self.name)

        cursor = self.connection.cursor()
        result = ()
        try:
            cursor.execute(sql, dataList)
            result = cursor.fetchall()
        except self._Warning:
            pass
        except self._OperationalError as e:
            # map to nive.utils.dataPool2.base.OperationalError
            raise OperationalError(e)
        #finally:
        #    cursor.close()
        return result

    
    def InsertFields(self, table, data, idColumn = None):
        """
        Insert row with multiple fields in the table.
        Codepage and dates are converted automatically
        Set `idColumn` to the column name of the auto increment unique id column
        to get it returned.
        
        returns the converted data 
        """
        dataList = []
        flds = []
        phdata = []
        ph = self.placeholder
        data = self.structure.serialize(table, None, data)
        for key, value in list(data.items()):
            flds.append(key)
            phdata.append(ph)
            dataList.append(value)

        sql = "INSERT INTO %s (%s) VALUES (%s)" % (table, ",".join(flds), ",".join(phdata))

        if self._debug:
            STACKF(0,sql+"\r\n",self._debug, self._log,name=self.name)

        cursor = self.connection.cursor()
        try:
            cursor.execute(sql, dataList)
        except self._Warning:
            pass
        except self._OperationalError as e:
            # map to nive.utils.dataPool2.base.OperationalError
            raise OperationalError(e)
        except:
            self.Undo()
            raise
        #finally:
        #    cursor.close()

        id = 0
        if idColumn:
            id = self._GetInsertIDValue(cursor)
        cursor.close()
        return data, id


    def UpdateFields(self, table, id, data, idColumn = "id", autoinsert=False):
        """
        Updates multiple fields in the table.
        If `autoinsert` is True the a new record is automatically inserted if it does not exist. Also
        the function returns the converted data *and* id (non zero if a new record was inserted)
        
        If `autoinsert` is False the function returns the converted data.
        """
        ph = self.placeholder
        if autoinsert:
            # if record does not exist, insert it
            sql = """select id from %s where %s=%s""" %(table, idColumn, ph)
            cursor = self.Execute(sql, (id,))
            r = cursor.fetchone()
            cursor.close()
            if not r:
                data, id = self.InsertFields(table, data, idColumn = idColumn)
                return data, id
            
        dataList = []
        data = self.structure.serialize(table, None, data)
        sql = ["UPDATE %s SET " % (table)]
        for key, value in list(data.items()):
            dataList.append(value)
            if len(sql)>1:
                sql.append(",%s=%s"%(key, ph))
            else:
                sql.append("%s=%s"%(key, ph))

        sql.append(" WHERE %s=%s" % (idColumn, ph))
        dataList.append(id)
        sql = "".join(sql)
        if self._debug:
            STACKF(0,sql+"\r\n",self._debug, self._log,name=self.name)

        cursor = self.connection.cursor()
        try:
            cursor.execute(sql, dataList)
        except self._Warning:
            pass
        except self._OperationalError as e:
            # map to nive.utils.dataPool2.base.OperationalError
            raise OperationalError(e)
        except:
            self.Undo()
            raise
        finally:
            cursor.close()

        if autoinsert:
            return data, 0
        return data


    def DeleteRecords(self, table, parameter):
        """
        Delete records referenced by parameters
        """
        if not parameter:
            return False
        p = []
        v = []
        ph = self.placeholder
        for field, value in list(parameter.items()):
            p.append("%s=%s"%(field, ph))
            v.append(value)
        sql = "DELETE FROM %s WHERE %s" % (table, " AND ".join(p))
        if self._debug:
            STACKF(0,sql+"\r\n\r\n",self._debug, self._log,name=self.name)
        cursor = self.connection.cursor()
        try:
            cursor.execute(sql, v)
        except:
            self.Undo()
            raise
        finally:
            cursor.close()


    # Text conversion -----------------------------------------------------------------------

    def EncodeText(self, text):
        """
        Used for text read from the database. 
        Converts the text to unicode based on self.dbCodePage.
        """
        if text is None:
            return text
        if isinstance(text, bytes):
            return str(text, self.dbCodePage, "replace")
        return text


    def DecodeText(self, text):
        """
        Used for text stored in database.
        Convert the text to unicode based on self.codePage.
        """
        if text is None:
            return text
        if isinstance(text, bytes):
            return str(text, self.codePage, "replace")
        return text


    def ConvertRecToDict(self, rec, flds):
        """
        Convert a database record tuple to dictionary based on flds list
        """
        return dict(list(zip(flds, rec)))


    # Entries -------------------------------------------------------------------------------------------

    def CreateEntry(self, pool_datatbl, id = 0, user = "", **kw):
        """
        Create new entry.
        Requires pool_datatbl as parameter
        """
        if id:
            id, dataref = self._CreateFixID(id, dataTbl=pool_datatbl)
        else:
            id, dataref = self._CreateNewID(table=self.MetaTable, dataTbl=pool_datatbl)
        if not id:
            return None
        kw["preload"] = "skip"
        kw["pool_dataref"] = dataref
        entry = self._GetPoolEntry(id, **kw)
        entry._InitNew(pool_datatbl, user)
        return entry


    def GetEntry(self, id, **kw):
        """
        Get entry from db by ID
        """
        return self._GetPoolEntry(id, **kw)


    def DeleteEntry(self, id, version=None):
        """
        Delete the entry and files
        """
        cursor = self.connection.cursor()

        # base record
        sql, values = self.FmtSQLSelect(["pool_dataref", "pool_datatbl"], parameter={"id":id}, dataTable=self.MetaTable, singleTable=1)
        if self._debug:
            STACKF(0,sql+"\r\n",self._debug, self._log,name=self.name)
        try:
            cursor.execute(sql, values)
        except:
            self.Undo()
            raise
        r = cursor.fetchone()
        if not r:
            return 0
        dataref = r[0]
        datatbl = r[1]
        cursor.close()

        self.DeleteFiles(id, version=version)
        tables = (self.MetaTable, self.FulltextTable, self.GroupsTable)
        for table in tables:
            self.DeleteRecords(table, parameter={"id":id})
        self.DeleteRecords(datatbl, parameter={"id":dataref})

        return 1


    def IsIDUsed(self, id):
        """
        Query database if id exists
        """
        sql, values = self.FmtSQLSelect(["id"], parameter={"id":id}, dataTable = self.MetaTable, singleTable=1)
        if self._debug:
            STACKF(0,sql+"\r\n",self._debug, self._log,name=self.name)
        c = self.connection.cursor()
        c.execute(sql, values)
        n = c.fetchone()
        c.close()
        return n is not None


    def GetBatch(self, ids, **kw):
        """
        Get all entries as objects at once. returns a list.
        supports preload: all, skip, meta
        
        kw: 
        - meta: list of meta pool_datatbl for faster lookup
        """
        preload = kw.get("preload", "all")
        entries = []
        version = kw.get("version")
        sort = kw.get("sort")
        if len(ids) == 0:
            return entries

        if preload == "skip":
            for id in ids:
                e = self._GetPoolEntry(id, preload="skip", version=version)
                entries.append(e)
            return entries

        if preload == "meta":
            flds = self.structure.get(self.MetaTable, version=version)
            if not flds:
                raise ConfigurationError("Meta layer is empty.")
            parameter = {"id": ids}
            operators = {"id": "IN"}
            sql, values = self.FmtSQLSelect(flds, parameter=parameter, dataTable=self.MetaTable, max=len(ids), sort=sort, operators=operators, singleTable=1)
            recs = self.Query(sql, values)
            for r in recs:
                meta = self.ConvertRecToDict(r, flds)
                e = self._GetPoolEntry(meta["id"], pool_dataref=meta["pool_dataref"], pool_datatbl=meta["pool_datatbl"], preload="skip")
                e._UpdateCache(meta = meta, data = None)
                entries.append(e)
            return entries

        if "meta" in kw and kw["meta"] and "pool_datatbl" in kw["meta"][0]:
            # use meta pool_datatbl passed in kw[meta]
            tables = []
            for r in kw["meta"]:
                if not r["pool_datatbl"] in tables:
                    tables.append(r["pool_datatbl"])
        else:
            # select data tables for ids
            flds = ["pool_datatbl"]
            parameter = {"id": ids}
            operators = {"id": "IN"}
            # select list of datatbls
            sql, values = self.FmtSQLSelect(flds, parameter=parameter, dataTable=self.MetaTable, sort=sort, groupby="pool_datatbl", operators=operators, singleTable=1)
            tables = []
            for r in self.Query(sql, values):
                if not r["pool_datatbl"] in tables:
                    tables.append(r[0])
        t = ""
        fldsm = self.structure.get(self.MetaTable, version=version)
        if not fldsm:
            raise ConfigurationError("Meta layer is empty.")
        unsorted = []
        for table in tables:
            structure = self.structure.get(table, version=version)
            if not structure:
                continue
            fldsd = list(structure)
            flds = list(fldsm) + fldsd
            parameter = {"id": ids, "pool_datatbl": table}
            operators = {"id": "IN", "pool_datatbl": "="}
            # select type data
            sql, values = self.FmtSQLSelect(flds, parameter=parameter, dataTable=table, sort=sort, operators=operators)
            typeData = self.Query(sql, values)
            for r2 in typeData:
                meta = self.ConvertRecToDict(r2[:len(fldsm)], fldsm)
                data = self.ConvertRecToDict(r2[len(fldsm):], fldsd)
                e = self._GetPoolEntry(meta["id"], pool_dataref=meta["pool_dataref"], pool_datatbl=meta["pool_datatbl"], preload="skip")
                meta = self.structure.deserialize(self.MetaTable, None, meta)
                data = self.structure.deserialize(table, None, data)
                e._UpdateCache(meta = meta, data = data)
                unsorted.append(e)
        # sort entries
        for id in ids:
            for o in unsorted:
                if o.id==id:
                    entries.append(o)
                    break
        return entries

    def _GetInsertIDValue(self, cursor):
        #("assert", "subclass")
        return 0

    def _CreateNewID(self, table="", dataTbl=None):
        #("assert", "subclass")
        return 0

    def _CreateFixID(self, id, dataTbl):
        #("assert", "subclass")
        return 0

    def _GetPoolEntry(self, id, **kw):
        #("assert", "subclass")
        return 0


    # Tree structure --------------------------------------------------------------

    def GetContainedIDs(self, base=0, sort="title", parameter=""):
        """
        Search subtree and returns a list of all contained ids.
        id needs to be first field, pool_unitref second
        """
        refs = """
        pool_unitref as ref1,
        (select pool_unitref from %(meta)s where id = ref1) as ref2,
        (select pool_unitref from %(meta)s where id = ref2) as ref3,
        (select pool_unitref from %(meta)s where id = ref3) as ref4,
        (select pool_unitref from %(meta)s where id = ref4) as ref5,
        (select pool_unitref from %(meta)s where id = ref5) as ref6,
        (select pool_unitref from %(meta)s where id = ref6) as ref7,
        (select pool_unitref from %(meta)s where id = ref7) as ref8,
        (select pool_unitref from %(meta)s where id = ref8) as ref9,
        (select pool_unitref from %(meta)s where id = ref9) as ref10,
        """ % {"meta": self.MetaTable}

        if parameter:
            parameter = "where " + parameter
        sql = """
        select %s id
        from %s %s
        order by pool_unitref, %s
        """ % (refs, self.MetaTable, parameter, sort)
        aC = self.connection.cursor()
        aC.execute(sql)
        l = aC.fetchall()
        aC.close()
        if len(l) == 0:
            return []

        ids = []
        for rec in l:
            #ignore outside tree
            if not base in rec or rec[10] == base:
                continue
            #data
            ids.append(rec[10])

        return ids


    def GetTree(self, flds=None, sort="title", base=0, parameter=""):
        """
        Loads a subtree as dictionary. 
        The returned dictionary has the following format:
        {"id": 123, "items": [{"id": 124, "items": [], "data": "q", ....}, ...], "data": "q", ....}
        
        id needs to be first field, pool_unitref second
        """
        if flds is None:
            flds = ["id"]
        refs = """
        pool_unitref as ref1,
        (select pool_unitref from %(meta)s where id = ref1) as ref2,
        (select pool_unitref from %(meta)s where id = ref2) as ref3,
        (select pool_unitref from %(meta)s where id = ref3) as ref4,
        (select pool_unitref from %(meta)s where id = ref4) as ref5,
        (select pool_unitref from %(meta)s where id = ref5) as ref6,
        (select pool_unitref from %(meta)s where id = ref6) as ref7,
        (select pool_unitref from %(meta)s where id = ref7) as ref8,
        (select pool_unitref from %(meta)s where id = ref8) as ref9,
        (select pool_unitref from %(meta)s where id = ref9) as ref10,
        """ % {"meta":self.MetaTable}

        if parameter:
            parameter = "where " + parameter
        sql = """
        select 
        %s 
        %s
        from %s %s
        order by pool_unitref, %s
        """ % (refs, ConvertListToStr(flds), self.MetaTable, parameter, sort)
        aC = self.connection.cursor()
        aC.execute(sql)
        l = aC.fetchall()
        tree = {"items":[]}
        aC.close()
        if len(l) == 0:
            return tree

        flds2 = ["ref1", "ref2", "ref3", "ref4", "ref5", "ref6", "ref7", "ref8", "ref9", "ref10"] + flds
        rtree = ["ref10", "ref9", "ref8", "ref7", "ref6", "ref5", "ref4", "ref3", "ref2", "ref1"]
        for rec in l:
            #ignore outside tree
            if not base in rec:  # or rec[10] == base:
                continue
            #data
            data = self.ConvertRecToDict(rec, flds2)
            #parent and path, lookup base in parents 
            add = 0
            current = tree
            for ref in rtree:
                if data[ref] is None:
                    continue
                if data[ref] == base:
                    add = 1
                if not add:
                    continue
                # add item to tree list
                if "items" not in current:
                    current["items"] = []
                refentry = self._InList(current["items"], data[ref])
                if not refentry:
                    refentry = {"id":data[ref], "items": []}
                    current["items"].append(refentry)
                current = refentry
            # add data 
            if "items" not in current:
                current["items"] = []
            entry = self._InList(current["items"], data["id"])
            if not entry:
                entry = data
                current["items"].append(entry)
            else:
                entry.update(data)

        return tree

    def _InList(self, l, id):
        for i in l:
            if i["id"] == id:
                return i
        return None
    
    
    def GetParentPath(self, id):
        """
        Returns id references of parents for the given id.
        Maximum 10 parents
        """
        if id <= 0:
            return []
        sql = """
        SELECT t1.pool_unitref AS ref1, 
         t2.pool_unitref as ref2, 
         t3.pool_unitref as ref3, 
         t4.pool_unitref as ref4,
         t5.pool_unitref as ref5,
         t6.pool_unitref as ref6,
         t7.pool_unitref as ref7,
         t8.pool_unitref as ref8,
         t9.pool_unitref as ref9,
         t10.pool_unitref as ref10
        FROM %(meta)s AS t1
        LEFT JOIN %(meta)s AS t2 ON t2.id = t1.pool_unitref
        LEFT JOIN %(meta)s AS t3 ON t3.id = t2.pool_unitref
        LEFT JOIN %(meta)s AS t4 ON t4.id = t3.pool_unitref
        LEFT JOIN %(meta)s AS t5 ON t5.id = t4.pool_unitref
        LEFT JOIN %(meta)s AS t6 ON t6.id = t5.pool_unitref
        LEFT JOIN %(meta)s AS t7 ON t7.id = t6.pool_unitref
        LEFT JOIN %(meta)s AS t8 ON t8.id = t7.pool_unitref
        LEFT JOIN %(meta)s AS t9 ON t9.id = t8.pool_unitref
        LEFT JOIN %(meta)s AS t10 ON t10.id = t9.pool_unitref
        WHERE t1.id = %(id)d;""" % {"id":id, "meta": self.MetaTable}
        
        if self._debug:
            STACKF(0,sql+"\r\n",self._debug, self._log,name=self.name)
        aC = self.connection.cursor()
        aC.execute(sql)
        aL = aC.fetchall()
        parents = []
        if len(aL) == 0:
            return parents
        cnt = 10
        for i in range(0, cnt-1):
            id = aL[0][i]
            if id == None:
                continue
            if id <= 0:
                break
            parents.insert(0, id)
        aC.close()
        return parents


    def GetParentTitles(self, id):
        """
        Returns titles of parents for the given id.
        maximum 10 parents
        """
        if id <= 0:
            return []
        aC = self.connection.cursor()
        sql = """
        SELECT t1.pool_unitref AS ref1, t1.title AS title1, 
         t2.pool_unitref as ref2, t2.title AS title2,
         t3.pool_unitref as ref3, t3.title AS title3,
         t4.pool_unitref as ref4, t4.title AS title4,
         t5.pool_unitref as ref5, t5.title AS title5,
         t6.pool_unitref as ref6, t6.title AS title6,
         t7.pool_unitref as ref7, t7.title AS title7,
         t8.pool_unitref as ref8, t8.title AS title8,
         t9.pool_unitref as ref9, t9.title AS title9,
         t10.pool_unitref as ref10, t10.title AS title10 
        FROM %(meta)s AS t1
        LEFT JOIN %(meta)s AS t2 ON t2.id = t1.pool_unitref
        LEFT JOIN %(meta)s AS t3 ON t3.id = t2.pool_unitref
        LEFT JOIN %(meta)s AS t4 ON t4.id = t3.pool_unitref
        LEFT JOIN %(meta)s AS t5 ON t5.id = t4.pool_unitref
        LEFT JOIN %(meta)s AS t6 ON t6.id = t5.pool_unitref
        LEFT JOIN %(meta)s AS t7 ON t7.id = t6.pool_unitref
        LEFT JOIN %(meta)s AS t8 ON t8.id = t7.pool_unitref
        LEFT JOIN %(meta)s AS t9 ON t9.id = t8.pool_unitref
        LEFT JOIN %(meta)s AS t10 ON t10.id = t9.pool_unitref
        WHERE t1.id = %(id)d;""" % {"id": id, "meta": self.MetaTable}        
        
        if self._debug:
            STACKF(0,sql+"\r\n",self._debug, self._log,name=self.name)
        aC.execute(sql)
        aL = aC.fetchall()
        parents = []
        if len(aL) == 0:
            return parents
        cnt = 10
        for i in range(1, cnt-1):
            title = aL[0][i*2+1]
            if title is None:
                break
            parents.insert(0, self.EncodeText(title))
        aC.close()
        return parents


    # Files and directories ---------------------------------------------------------------------

    def InitFileStorage(self, root, connectionParam):
        """
        Set the local root path for files
        """
        self.root = DvPath()
        self.root.SetStr(root)

    def GetRoot(self):
        return str(self.root)

    def DeleteFiles(self, id, version):
        # subclassed
        pass


    # groups - userid assignment storage ------------------------------------------------------------------------------------

    def GetGroups(self, id, userid=None, group=None):
        """
        Get local group assignment for userid.

        `id` can be a single value or a tuple. If its a tuple the groups for
        all matching ids are returned.

        returns a group assignment list [["userid", "groupid", "id"], ...]
        """
        # check if exists
        parameter = {}
        
        if id is not None:
            parameter["id"] = id
        else:
            raise TypeError("id must not be none")
        if userid:
            parameter["userid"] = userid
        if group:
            parameter["groupid"] = group
        operators = {}
        if isinstance(id, (list, tuple)):
            operators["id"] = "IN"
        sql, values = self.FmtSQLSelect(["userid", "groupid", "id"],
                                        parameter=parameter,
                                        operators=operators,
                                        dataTable=self.GroupsTable,
                                        singleTable=1)
        r = self.Query(sql, values)
        return r


    def AddGroup(self, id, userid, group):
        """
        Add a local group assignment for userid.
        """
        data = {"userid": userid, "groupid": group}
        if id is not None:
            data["id"] = id
        else:
            raise TypeError("id must not be none")
        self.InsertFields(self.GroupsTable, data)


    def RemoveGroups(self, id, userid=None, group=None):
        """
        Remove a local group assignment for userid or all for the id/ref.
        """
        p = {}
        if id is not None:
            p["id"] = id
        else:
            raise TypeError("id must not be none")
        if userid:
            p["userid"] = userid
        if group:
            p["groupid"] = group
        self.DeleteRecords(self.GroupsTable, p)


    def GetAllUserGroups(self, userid):
        """
        Get all local group assignment for userid.
        
        returns a group assignment list [["userid", "groupid", "id"], ...]
        """
        # check if exists
        p = {"userid": userid}
        o = {"userid": "="}
        sql, values = self.FmtSQLSelect(["userid", "groupid", "id"], parameter=p, operators=o, dataTable=self.GroupsTable, singleTable=1)
        r = self.Query(sql, values)
        return r


    # Pool status ---------------------------------------------------------------------

    def GetCountEntries(self, table = None):
        """
        Returns the total number of entries in the pool
        """
        if not table:
            table = self.MetaTable
        aC = self.connection.cursor()
        aC.execute("SELECT COUNT(*) FROM %s" % (table))
        aN = aC.fetchone()[0]
        aC.close()
        return aN


    def SetConnection(self, conn):
        self._conn = conn

    def CreateConnection(self, connParam):
        return self._DefaultConnection(connParam)

    
    # internal subclassing -------------------------------------------
    
    def _GetDefaultPoolStructure(self):
        return PoolStructure

    def _GetDataWrapper(self):
        return DataWrapper

    def _GetMetaWrapper(self):
        return MetaWrapper

    def _GetFileWrapper(self):
        return FileWrapper
    
    # to be removed --------------------------------------------------
    
    def GetConnection(self):
        return self._conn




class Entry(object):
    """
    Entry object of data pool
    """

    def __init__(self, dataPool, id, **kw):
        """[a]
        version: version of this entry
        dataTbl: data table name of this entry type
        preload: meta/data flds to preload on init. "all" (default), "skip", "stdmetadata", "stdmeta", "meta"
        kw virtual = 1 to disable preload() and exists() 
        """
        # key values
        self.id = id
        self.version = kw.get("version", None)
        self.dataTbl = kw.get("pool_datatbl", None)
        self.dataRef = kw.get("pool_dataref", 0)

        # basic properties
        self.pool = dataPool
        selfref = weakref.ref(self)
        self.meta = dataPool._GetMetaWrapper()(selfref)
        self.data = dataPool._GetDataWrapper()(selfref)
        self.files = dataPool._GetFileWrapper()(selfref)
        self.cacheRec = True
        self._localRoles = None
        self._permissions = None
        self.virtual = kw.get("virtual")==1
        
        #self._debug = 0#dataPool._debug

        if self.virtual:
            return
        preload = kw.get("preload", "all")
        if self.id and self.cacheRec and preload != "skip":
            self.Load(preload)
            if self.data.IsEmpty() and self.meta.IsEmpty() and self.files.IsEmpty():
                if not self.Exists():
                    raise NotFound("%s not found" % str(id))
        #else:
        #    if not self.Exists():
        #        raise NotFound, "%s not found" % str(id)


    def __del__(self):
        self.Close()


    def Close(self):
        self.meta.close()
        self.data.close()
        self.files.close()
        self.pool = None

    def Exists(self):
        """
        Check if the entry physically exists in the database
        """
        if self.virtual:
            return True
        if not self.IsValid():
            return False
        return self.pool.IsIDUsed(self.id)

    def IsValid(self):
        """
        Check if the id is valid
        """
        return self.id > 0


    def SerializeValue(self, fieldname, value, meta=False):
        if meta:
            tbl = self.pool.MetaTable
        else:
            tbl = self.GetDataTbl()
        return self.pool.structure.serialize(tbl, fieldname, value)

    def DeserializeValue(self, fieldname, value, meta=False):
        if meta:
            tbl = self.pool.MetaTable
        else:
            tbl = self.GetDataTbl()
        return self.pool.structure.deserialize(tbl, fieldname, value)

    # Transactions ------------------------------------------------------------------------

    def Commit(self, user="", dbCommit=True):
        """
        Commit temporary changes (meta, data, files) to database
        """
        self.Touch(user)
        try:
            # meta
            if self.meta.HasTemp():
                self.pool.UpdateFields(self.pool.MetaTable, self.id, self.meta.GetTemp())
            # data
            if self.data.HasTemp():
                self.pool.UpdateFields(self.GetDataTbl(), self.GetDataRef(), self.data.GetTemp())
            # files
            if self.files.HasTemp():
                self.CommitFiles(self.files.GetTemp())
            if dbCommit:
                self.pool.Commit()
            # remove previous files
            self.Cleanup(self.files.GetTemp())
            self.data.SetContent(self.data.GetTemp())
            self.data.clear()
            self.meta.SetContent(self.meta.GetTemp())
            self.meta.clear()
            self.files.SetContent(self.files.GetTemp())
            self.files.clear()
        except Exception as e:
            try:
                self.Undo()
            except:
                pass
            raise 
        return True


    def Undo(self):
        """
        Undo changes in database
        """
        self.meta.EmptyTemp()
        self.data.EmptyTemp()
        self.files.EmptyTemp()
        self.pool.Undo()


    # Reading Values ------------------------------------------------------------------------

    def GetMetaField(self, fld, fromDB=False):
        """
        Read a single field from meta layer
        if fromDB data is loaded from db, not cache
        """
        if fld == "id":
            return self.id
        elif fld == "pool_size":
            return self.GetDataSize()

        if not fromDB and self.cacheRec and fld in self.meta:
            return self.meta[fld]

        sql, values = self.pool.FmtSQLSelect([fld], parameter={"id":self.id}, dataTable=self.pool.MetaTable, singleTable=1)
        data = self._GetFld(sql, values)
        data = self.pool.structure.deserialize(self.pool.MetaTable, fld, data)
        self._UpdateCache({fld: data})
        return data


    def GetDataField(self, fld, fromDB=False):
        """
        Read a single field from data layer
        if fromDB data is loaded from db, not cache
        """
        if not fromDB and self.cacheRec and fld in self.data:
            return self.data[fld]

        tbl = self.GetDataTbl()
        sql, values = self.pool.FmtSQLSelect([fld], parameter={"id":self.GetDataRef()}, dataTable=tbl, singleTable=1)
        data = self._GetFld(sql, values)
        data = self.pool.structure.deserialize(tbl, fld, data)
        self._UpdateCache(None, {fld: data})
        return data


    def GetMeta(self):
        """
        Read meta layer and return as dictionary
        """
        if self.cacheRec and not self.meta.IsEmpty():
            return self.meta.copy()

        data =  self._SQLSelect(self.pool.structure.get(self.pool.MetaTable, version=self.version))
        data = self.pool.structure.deserialize(self.pool.MetaTable, None, data)
        self._UpdateCache(data)
        data["id"] = self.id
        return data


    def GetData(self):
        """
        Read data layer and return as dictionary
        """
        if self.cacheRec and not self.data.IsEmpty():
            return self.data

        tbl = self.GetDataTbl()
        try:
            flds = self.pool.structure.get(tbl, version=self.version)
        except:
            return None

        data = self._SQLSelect(flds, table=tbl)
        data = self.pool.structure.deserialize(tbl, None, data)
        self._UpdateCache(None, data)
        return data


    # Writing Values ------------------------------------------------------------------------

    def SetMetaField(self, fld, data, cache=True):
        """
        Update single field to meta layer.
        Commits changes immediately to database without calling touch.
        """
        temp = self.pool.UpdateFields(self.pool.MetaTable, self.id, {fld:data})
        self.pool.Commit()
        if cache:
            self._UpdateCache(temp)
        return True


    def SetDataField(self, fld, data, cache=True):
        """
        Update single field to data layer
        Commits changes immediately to database without calling touch.
        """
        if fld == "id":
            return False

        # check if data record already exists
        id = self.GetDataRef()
        if id <= 0:
            return False

        temp = self.pool.UpdateFields(self.GetDataTbl(), id, {fld:data})
        self.pool.Commit()

        if cache:
            self._UpdateCache(None, temp)
        return True


    def Touch(self, user=""):
        """
        Tets change date and changed by to now
        """
        temp = {}
        temp["pool_change"] = self.pool.GetDBDate()
        if user is None:
            user=""
        else:
            user=str(user)
        temp["pool_changedby"] = user
        self.meta.update(temp)
        return True


    # Files -------------------------------------------------------------------------------------------

    def DuplicateFiles(self, newEntry):
        """
        """
        #BREAK("subclass")
        pass


    def CommitFile(self, key, file):
        """
        """
        #BREAK("subclass")
        pass


    # Actions ---------------------------------------------------------------------------

    def Duplicate(self, duplicateFiles = True):
        """
        Create a copy of the entry with a new id.
        """
        if self.cacheRec:
            self.Load("all")

        newEntry = self.pool.CreateEntry(pool_datatbl=self.GetDataTbl())
        if not newEntry:
            return None
        id = newEntry.GetID()

        # copy data
        newEntry.meta.update(self.GetMeta())
        newEntry.data.update(self.GetData())

        # check if entry contains file and file exists
        if duplicateFiles:
            if not self.DuplicateFiles(newEntry):
                self.pool.DeleteEntry(id)
                del newEntry
                return None

        return newEntry


    # Preloading -------------------------------------------------------------------------------------------

    def Load(self, option = "all", reload = False):
        """
        Loads different sets of fields in single sql statement

        options:
        skip
        all: all meta and data fields
        stdmeta: fields configured in stdMeta list
        stdmetadata: fields configured in stdMeta list and all data fields
        meta: all meta fields
        """
        if self.virtual:
            return True
        if option == "skip":
            return True

        # check data already in memory
        if not reload:
            if option == "all" and not self.data.IsEmpty() and not self.meta.IsEmpty():
                return True

            elif option == "stdmetadata" and not self.data.IsEmpty() and not self.meta.IsEmpty():
                return True

            elif option == "stdmeta" and not self.meta.IsEmpty():
                return True

            elif option == "meta" and not self.meta.IsEmpty():
                return True

        # shortcut sql load functions based on pool structure
        if option == "all":
            return self._PreloadAll()
        elif option == "stdmetadata":
            return self._PreloadStdMetaData()
        elif option == "stdmeta":
            return self._PreloadStdMeta()
        elif option == "meta":
            return self._PreloadMeta()

        return True


    # System Values ------------------------------------------------------------------------

    def GetID(self):
        return self.id

    def GetDataTbl(self):
        if self.dataTbl:
            return self.dataTbl
        self.dataTbl = self.GetMetaField("pool_datatbl")
        return self.dataTbl

    def GetDataRef(self):
        if self.dataRef:
            return self.dataRef
        self.dataRef = self.GetMetaField("pool_dataref")
        return self.dataRef

    def GetSize(self):
        return 0

    def GetVersion(self):
        return self.version


    # Caching --------------------------------------------------------------------

    def Clear(self):
        """
        empty cache
        """
        self.meta.clear()
        self.data.clear()
        self.files.clear()


    # Fulltext ---------------------------------------------------------------------------------------

    def WriteFulltext(self, text):
        """
        Update or create fulltext for entry
        """
        if text is None:
            text=""
        sql, values = self.pool.FmtSQLSelect(["id"], parameter={"id":self.id}, dataTable=self.pool.FulltextTable, singleTable=1)
        cursor = self.pool.connection.cursor()
        if self.pool._debug:
            STACKF(0,sql+"\r\n\r\n", self.pool._debug, self.pool._log, name=self.pool.name)
        try:
            cursor.execute(sql, values)
        except:
            self.pool.Undo()
            raise
        r = cursor.fetchone()
        text = self.pool.DecodeText(text)
        if not r:
            self.pool.InsertFields(self.pool.FulltextTable, {"text": text, "id": self.id})
        else:
            self.pool.UpdateFields(self.pool.FulltextTable, self.id, {"text": text})
        cursor.close()


    def GetFulltext(self):
        """
        read fulltext from entry
        """
        sql, values = self.pool.FmtSQLSelect(["text"], parameter={"id":self.id}, dataTable=self.pool.FulltextTable, singleTable=1)
        cursor = self.pool.connection.cursor()
        if self.pool._debug:
            STACKF(0,sql+"\r\n\r\n", self.pool._debug, self.pool._log, name=self.pool.name)
        try:
            cursor.execute(sql, values)
        except:
            self.pool.Undo()
            raise
        r = cursor.fetchone()
        cursor.close()
        if not r:
            return ""
        return self.pool.DecodeText(r[0])


    def DeleteFulltext(self):
        """
        Delete fulltext for entry
        """
        ph = self.pool.placeholder
        sql = "DELETE FROM %s WHERE id = %s"%(self.pool.FulltextTable, ph)
        if self.pool._debug:
            STACKF(0,sql+"\r\n\r\n",self.pool._debug, self.pool._log,name=self.pool.name)
        aCursor = self.pool.connection.cursor()
        aCursor.execute(sql, (self.id,))
        aCursor.close()


    # Locking ---------------------------------------------------------------------

    def Lock(self, owner):
        return False

    def UnLock(self, owner, forceUnlock = False):
        return False

    def IsLocked(self, owner):
        return False


    # SQL functions -------------------------------------------------------------------------------------------

    def _GetFld(self, sql, values):
        """
        select single field from sql
        """
        try:
            cursor = self.pool.connection.cursor()
            if self.pool._debug:
                STACKF(0,sql+"\r\n",self.pool._debug, self.pool._log,name=self.pool.name)
            cursor.execute(sql, values)
            r = cursor.fetchone()
            cursor.close()
            return r[0]
        except self.pool._Warning as e:
            try:    cursor.close()
            except: pass
            if self.pool._debug:
                STACKF(0,str(e)+"\r\n",self.pool._debug, self.pool._log,name=self.pool.name)
            return None
        except:
            self.pool.Undo()
            raise
    
    
    def _SQLSelect(self, flds, cursor=None, table=None):
        """
        select one entry and convert to dictionary
        """
        pool = self.pool
        if not table:
            table = pool.MetaTable
            param = {"id": self.id}
        else:
            param = {"id": self.GetDataRef()}
        sql, values = pool.FmtSQLSelect(flds, param, dataTable = table, version=self.version, singleTable=1)
        c = 0
        if cursor is None:
            c = 1
            cursor = pool.connection.cursor()
        if pool._debug:
            STACKF(0,sql+"\r\n\r\n",pool._debug, pool._log,name=pool.name)
        try:
            cursor.execute(sql, values)
        except:
            pool.Undo()
            raise
        r = cursor.fetchone()
        if c:
            cursor.close()
        if not r:
            return None
        return pool.ConvertRecToDict(r, flds)


    def _SQLSelectAll(self, metaFlds, dataFlds, cursor=None):
        """
        select meta and data of one entry and convert to dictionary
        """
        pool = self.pool
        flds2 = list(metaFlds) + list(dataFlds)
        param = {"id": self.id}
        sql, values = pool.FmtSQLSelect(flds2, param, dataTable = self.GetDataTbl(), version=self.version)

        c = 0
        if cursor is None:
            c = 1
            cursor = pool.connection.cursor()
        if self.pool._debug:
            STACKF(0,sql+"\r\n\r\n",pool._debug, pool._log,name=pool.name)
        try:
            cursor.execute(sql, values)
        except:
            pool.Undo()
            raise
        r = cursor.fetchone()
        if c:
            cursor.close()
        if not r:
            raise TypeError(sql)
        # split data and meta
        meta = pool.ConvertRecToDict(r[:len(metaFlds)], metaFlds)
        data = pool.ConvertRecToDict(r[len(metaFlds):], dataFlds)
        return meta, data


    def _PreloadAll(self, cursor=None):
        if self.virtual:
            return True
        if not self.dataTbl or not self.dataRef:
            if not self._PreloadMeta(cursor):
                return False
            self._PreloadData(cursor)
            return True
        dataTbl = self.GetDataTbl()
        meta, data = self._SQLSelectAll(self.pool.structure.get(self.pool.MetaTable, version=self.version),
                                        self.pool.structure.get(dataTbl, version=self.version), cursor=cursor)
        if not meta:
            return False
        meta = self.pool.structure.deserialize(self.pool.MetaTable, None, meta)
        data = self.pool.structure.deserialize(dataTbl, None, data)
        self._UpdateCache(meta, data)
        return True


    def _PreloadMeta(self, cursor=None):
        if self.virtual:
            return True
        meta = self._SQLSelect(self.pool.structure.get(self.pool.MetaTable, self.version), cursor=cursor)
        if not meta:
            return False
        meta = self.pool.structure.deserialize(self.pool.MetaTable, None, meta)
        self._UpdateCache(meta)
        return True


    def _PreloadData(self, cursor=None):
        if self.virtual:
            return True
        dataTbl = self.GetDataTbl()
        data = self._SQLSelect(self.pool.structure.get(dataTbl, version=self.version), cursor=cursor, table=dataTbl)
        if not data:
            return False
        data = self.pool.structure.deserialize(dataTbl, None, data)
        self._UpdateCache(None, data)
        return True


    def _PreloadStdMeta(self, cursor=None):
        if self.virtual:
            return True
        meta = self._SQLSelect(self.pool.structure.stdMeta, cursor=cursor)
        if not meta:
            return False
        meta = self.pool.structure.deserialize(self.pool.MetaTable, None, meta)
        self._UpdateCache(meta)
        return True


    def _PreloadStdMetaData(self, cursor=None):
        if self.virtual:
            return True
        dataTbl = self.GetDataTbl()
        meta, data = self._SQLSelectAll(self.pool.structure.stdMeta, self.pool.structure.get(dataTbl, version=self.version), cursor=cursor)
        if not meta:
            return False
        meta = self.pool.structure.deserialize(self.pool.MetaTable, None, meta)
        data = self.pool.structure.deserialize(dataTbl, None, data)
        self._UpdateCache(meta, data)
        return True


    # Internal -------------------------------------------------------------------------------------------

    def _InitNew(self, dataTable = None, user = ""):
        """
        sets create date and created by to now
        """
        aC = self.pool.connection.cursor()
        self.dataTbl = dataTable
        aMeta = {}
        date = self.pool.GetDBDate()
        aMeta["pool_create"] = date
        aMeta["pool_change"] = date
        if user is None:
            user=""
        else:
            user=str(user)
        aMeta["pool_createdby"] = user
        aMeta["pool_changedby"] = user
        if self.dataTbl and self.dataTbl != "":
            aMeta["pool_datatbl"] = self.dataTbl
        self.meta.update(aMeta, force=True)


    def _UpdateCache(self, meta = None, data = None, files = None):
        if meta:
            self.meta.SetContent(meta)
        if data:
            self.data.SetContent(data)
        if files:
            self.files.SetContent(files)



class NotFound(Exception):
    """ raised if entry not found """

class FileNotFound(Exception):
    """ raised if physical file not found """




import csv
import random
import ntpath
import io
import copy
import base64
import os
import datetime

from nive.i18n import _
from nive.definitions import Conf
from nive.components.reform import schema
from nive.components.reform.schema import Invalid
from nive.components.reform.schema import null
from nive.helper import File

try:
    import json 
except ImportError: # PRAGMA: no cover
    import simplejson as json 


class Widget(object):
    """
    A widget is the building block for rendering logic.  The
    :class:`reform.widget.Widget` class is never instantiated
    directly: it is the abstract class from which all other widget
    types within :mod:`reform.widget` derive.  It should likely also
    be subclassed by application-developer-defined widgets.

    A widget instance is attached to a field during normal operation.
    A widget is not meant to carry any state.  Instead, widget
    implementations should use the ``field`` object passed to them
    during :meth:`reform.widget.Widget.serialize` and
    :meth:`reform.widget.Widget.deserialize` as a scratchpad for state
    information.

    All widgets have the following attributes:

    hidden
        An attribute indicating the hidden state of this widget.  The
        default is ``False``.  If this attribute is not ``False``, the
        field associated with this widget will not be rendered in the
        form (although, if the widget is a structural widget, its
        children will be; ``hidden`` is not a recursive flag).  No
        label, no error message, nor any furniture such as a close
        button when the widget is one of a sequence will exist for the
        field in the rendered form.

    category
        A string value indicating the *category* of this widget.  This
        attribute exists to inform structural widget rendering
        behavior.  For example, when a text widget or another simple
        'leaf-level' widget is rendered as a child of a mapping widget
        using the default template mapping template, the field title
        associated with the child widget will be rendered above the
        field as a label by default.  This is because simple child
        widgets are in the ``default`` category and no special action
        is taken when a structural widget renders child widgets that
        are in the ``default`` category.  However, if the default
        mapping widget encounters a child widget with the category of
        ``structural`` during rendering (the default mapping and
        sequence widgets are in this category), it omits the title.
        Default: ``default``

    error_class
        The name of the CSS class attached to various tags in the form
        renderering indicating an error condition for the field
        associated with this widget.  Default: ``error``.
    
    css_class
        The name of the CSS class attached to various tags in
        the form renderering specifying a new class for the field
        associated with this widget.  Default: ``None`` (no class).

    requirements
        A sequence of two-tuples in the form ``( (requirement_name,
        version_id), ...)`` indicating the logical external
        requirements needed to make this widget render properly within
        a form.  The ``requirement_name`` is a string that *logically*
        (not concretely, it is not a filename) identifies *one or
        more* Javascript or CSS resources that must be included in the
        page by the application performing the form rendering.  The
        requirement name string value should be interpreted as a
        logical requirement name (e.g. ``jquery`` for JQuery,
        'tinymce' for Tiny MCE).  The ``version_id`` is a string
        indicating the version number (or ``None`` if no particular
        version is required).  For example, a rich text widget might
        declare ``requirements = (('tinymce', '3.3.8'),)``.  See also:
        :ref:`specifying_widget_requirements` and
        :ref:`widget_requirements`.

        Default: ``()`` (the empty tuple, meaning no special
        requirements).
        
    configuration
        A ``nive.definitions.FieldConf`` instance or dictionary 
        providing configuration values used to render the widget.
        Values depend on the widget type if used at all. Used 
        for example to customize Tiny MCE or other javascript modules.
        See the widgets documentation on how to use the configuration.

    These attributes are also accepted as keyword arguments to all
    widget constructors; if they are passed, they will override the
    defaults.

    Particular widget types also accept other keyword arguments that
    get attached to the widget as attributes.  These are documented as
    'Attributes/Arguments' within the documentation of each concrete
    widget implementation subclass.
    """

    hidden = False
    category = 'default'
    error_class = 'error'
    css_class = 'form-control'
    requirements = ()
    configuration = None
    # test support
    form = Conf(widget=Conf(settings={}))

    def __init__(self, **kw):
        self.__dict__.update(kw)
        self.settings = {}


    def serialize(self, field, cstruct):
        """
        The ``serialize`` method of a widget must serialize a
        :term:`cstruct` value to an HTML rendering.  A :term:`cstruct`
        value is the value which results from a :term:`Colander`
        schema serialization for the schema node associated with this
        widget.  ``serialize`` should return the HTML rendering: the
        result of this method should always be a string containing
        HTML.  The ``field`` argument is the :term:`field` object to
        which this widget is attached.  
        """
        raise NotImplementedError

    def deserialize(self, field, pstruct, formstruct=None):
        """
        The ``deserialize`` method of a widget must deserialize a
        :term:`pstruct` value to a :term:`cstruct` value and return the
        :term:`cstruct` value.  The ``pstruct`` argument is a value looked up in
        the request or source data. ``formstruct`` contains all values. The
        ``field`` argument is the field object to which this widget is
        attached.
        """
        raise NotImplementedError

    def handle_error(self, field, error):
        """
        The ``handle_error`` method of a widget must:

        - Set the ``error`` attribute of the ``field`` object it is
          passed, if the ``error`` attribute has not already been set.

        - Call the ``handle_error`` method of each subfield which also
          has an error (as per the ``error`` argument's ``children``
          attribute).
        """
        if field.error is None:
            field.error = error
        # XXX exponential time
        for e in error.children:
            for num, subfield in enumerate(field.children):
                if e.pos == num:
                    subfield.widget.handle_error(subfield, e)


class FormWidget(Widget):
    """
    The top-level widget; represents an entire form.

    **Attributes/Arguments**

    template
        The template name used to render the widget.  Default:
        ``form``.

    item_template
        The template name used to render each item in the form.
        Default: ``mapping_item``.

    """
    template = 'form'
    item_template = 'field'
    action_template = 'form_actions'
    error_class = 'error'
    category = 'structural'
    requirements = ( ('reform', None), )

    def serialize(self, field, cstruct, **kw):
        if cstruct in (null, None):
            cstruct = {}
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct,
                              null=null, **kw)

    def deserialize(self, field, pstruct, formstruct=None):
        error = None
        
        result = {}

        if pstruct is null:
            pstruct = {}

        for num, subfield in enumerate(field.children):
            name = subfield.name
            subval = pstruct.get(name, null)
                            
            try:
                result[name] = subfield.deserialize(subval, pstruct)
            except Invalid as e:
                result[name] = e.value
                if error is None:
                    error = Invalid(field.schema, value=result)
                error.add(e, num)

        if error is not None:
            raise error

        return result
    
    
class TextInputWidget(Widget):
    """
    Renders an ``<input type="text">`` widget.

    **Attributes/Arguments**

    size
        The size, in columns, of the text input field.  Defaults to
        ``None``, meaning that the ``size`` is not included in the
        widget output (uses browser default size).

    input_type
        Html input.type value. Default = text

    template
       The template name used to render the widget.  Default:
        ``textinput``.

    strip
        If true, during deserialization, strip the value of leading
        and trailing whitespace (default ``True``).

    mask
        A :term:`jquery.maskedinput` input mask, as a string.

        a - Represents an alpha character (A-Z,a-z)
        9 - Represents a numeric character (0-9)
        * - Represents an alphanumeric character (A-Z,a-z,0-9)

        All other characters in the mask will be considered mask
        literals.

        Example masks:

          Date: 99/99/9999

          US Phone: (999) 999-9999

          US SSN: 999-99-9999

        When this option is used, the :term:`jquery.maskedinput`
        library must be loaded into the page serving the form for the
        mask argument to have any effect.  See :ref:`masked_input`.

    mask_placeholder
        The placeholder for required nonliteral elements when a mask
        is used.  Default: ``_`` (underscore).
    """
    template = 'textinput'
    input_type = None
    size = None
    strip = True
    mask = None
    mask_placeholder = "_"
    placeholder = ""
    requirements = ( ('jquery.maskedinput', None), )

    def serialize(self, field, cstruct):
        if cstruct in (null, None):
            cstruct = ''
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct is null:
            return null
        if self.strip:
            pstruct = pstruct.strip()
        if not pstruct and field.required:
            return null
        return pstruct

class AutocompleteInputWidget(Widget):
    """
    Renders an ``<input type="text">`` widget which provides
    autocompletion via a list of values.

    When this option is used, the :term:`jquery.ui.autocomplete`
    library must be loaded into the page serving the form for
    autocompletion to have any effect.  See also
    :ref:`autocomplete_input`.  A version of :term:`JQuery UI` which
    includes the autoinclude sublibrary is included in the reform
    static directory. The default styles for JQuery UI are also
    available in the reform static/css directory.

    **Attributes/Arguments**

    size
        The size, in columns, of the text input field.  Defaults to
        ``None``, meaning that the ``size`` is not included in the
        widget output (uses browser default size).

    template
        The template name used to render the widget.  Default:
        ``autocomplete_textinput``.

    strip
        If true, during deserialization, strip the value of leading
        and trailing whitespace (default ``True``).

    values
        ``values`` from which :term:`jquery.ui.autocomplete` provides
        autocompletion. It MUST be an iterable that can be converted
        to a json array by [simple]json.dumps. It is also possible
        to pass a [base]string representing a remote URL.

        If ``values`` is a string it will be treated as a
        URL. If values is an iterable which can be serialized to a
        :term:`json` array, it will be treated as local data.

        If a string is provided to a URL, an :term:`xhr` request will
        be sent to the URL. The response should be a JSON
        serialization of a list of values.  For example:

          ['foo', 'bar', 'baz']

        Defaults to ``None``.

    min_length
        ``min_length`` is an optional argument to
        :term:`jquery.ui.autocomplete`. The number of characters to
        wait for before activating the autocomplete call.  Defaults to
        ``2``.

    delay
        ``delay`` is an optional argument to
        :term:`jquery.ui.autocomplete`. It sets the time to wait after a
        keypress to activate the autocomplete call.
        Defaults to ``10`` ms or ``400`` ms if a url is passed.
    """
    delay = None
    min_length = 2
    size = None
    strip = True
    template = 'autocomplete_input'
    values = None
    requirements = ( ('jqueryui', None), )

    def serialize(self, field, cstruct):
        if cstruct in (null, None):
            cstruct = ''
        options = {}
        if not self.delay:
            # set default delay if None
            options['delay'] = (isinstance(self.values,
                                          str) and 400) or 10
        options['minLength'] = self.min_length
        options = json.dumps(options)
        values = json.dumps(self.values)
        template = self.template
        return field.renderer(template,
                              cstruct=cstruct,
                              field=field,
                              options=options,
                              values=values)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct is null:
            return null
        if self.strip:
            pstruct = pstruct.strip()
        if not pstruct:
            return null
        return pstruct


class DateInputWidget(Widget):
    """
    
    Renders a JQuery UI date picker widget
    (http://jqueryui.com/demos/datepicker/).  Most useful when the
    schema node is a ``colander.Date`` object.

    **Attributes/Arguments**

    size
        The size, in columns, of the text input field.  Defaults to
        ``None``, meaning that the ``size`` is not included in the
        widget output (uses browser default size).

    template
        The template name used to render the widget.  Default:
        ``dateinput``.

    dateFormat
        A list of allowed date formats. Used to parsed the input.

    """
    template = 'dateinput'
    size = None
    requirements = ( )
    dateFormat = ("%Y-%m-%d","%Y/%m/%d","%m/%d/%y","%m/%d/%Y")

    def serialize(self, field, cstruct):
        if cstruct in (null, None):
            cstruct = ''
        else:
            try:
                dt = datetime.datetime.strptime(cstruct,"%Y-%m-%d")
                cstruct =  dt.strftime(self.dateFormat[0])
            except ValueError:
                # keep the date string as it is
                pass
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct in ('', null):
            return null
        pstruct = pstruct.replace(" ", "")
        for fmt in self.dateFormat:
            try:
                dt = datetime.datetime.strptime(pstruct,fmt)
                # use the current year if not given. supports formats like '%d.%m'
                if dt.year==1900 and not "1900" in pstruct:
                    pstruct = str(datetime.datetime.now().year) + dt.strftime("-%m-%d")
                else:
                    pstruct = dt.strftime("%Y-%m-%d")
                break
            except ValueError:
                pass
        return pstruct

class DateTimeInputWidget(DateInputWidget):
    """
    Renders a date picker with a JQuery Plugin tempusdominus add-on.
    Used for ``colander.DateTime`` schema nodes.

    **Attributes/Arguments**

    options
        A dictionary of options that's passed to the datetimepicker.

    size
        The size, in columns, of the text input field.  Defaults to
        ``None``, meaning that the ``size`` is not included in the
        widget output (uses browser default size).

    template
        The template name used to render the widget.  Default:
        ``dateinput``.

    """
    template = 'datetimeinput'
    size = None
    requirements = ( ('datetimepicker', None), )
    option_defaults = {'dateFormat': '%Y-%m-%d',
                       'timeFormat': '%H:%M',
                       'separator': ' '}
    options = {}

    def _options(self):
        options = self.option_defaults.copy()
        options.update(self.options)
        return options

    def serialize(self, field, cstruct):
        if cstruct in (null, None):
            cstruct = ''
        template = self.template
        options = self._options()
        if len(cstruct) == 25: # strip timezone if it's there
            cstruct = cstruct[:-6]

        temp = dict(date='', time='')
        opt = self._options()
        if cstruct:
            try:
                dt = datetime.datetime.strptime(cstruct, "%Y-%m-%dT%H:%M:%S")
                if cstruct:
                    temp['date'] = dt.strftime(opt['dateFormat'])
                    temp['time'] = dt.strftime(opt['timeFormat'])
                    test = datetime.datetime(year=2000,day=31,month=12)
                    if temp['time']==test.strftime(opt['timeFormat']):
                        temp['time'] = ''
            except ValueError:
                try:
                    temp['date'], temp['time'] = cstruct.split(opt['separator'])
                except ValueError:
                    temp['date'] = field.widget.form.view.GetFormValue(field.name+'-dt')
                    temp['time'] = field.widget.form.view.GetFormValue(field.name+'-tm')

        cstruct = temp

        return field.renderer(
            template,
            field=field,
            cstruct=cstruct,
            options=json.dumps(opt),
            )

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct in (null,):
            if not field.required and pstruct=='':
                # empty value allowed
                return ''
            return null
        pstruct = pstruct.strip()
        if pstruct in ('',):
            if not field.required and pstruct=='':
                # empty value allowed
                return ''
            return null
        opt = self._options()
        try:
            dt = datetime.datetime.strptime(pstruct, opt['dateFormat'] + opt['separator'] + opt['timeFormat'])
            pstruct = dt.strftime("%Y-%m-%dT%H:%M:%S")
        except ValueError:
            pass
        return pstruct

class TextAreaWidget(TextInputWidget):
    """
    Renders a ``<textarea>`` widget.

    **Attributes/Arguments**

    cols
        The size, in columns, of the text input field.  Defaults to
        ``None``, meaning that the ``cols`` is not included in the
        widget output (uses browser default cols).

    rows
        The size, in rows, of the text input field.  Defaults to
        ``None``, meaning that the ``rows`` is not included in the
        widget output (uses browser default cols).

    template
        The template name used to render the widget.  Default:
        ``textarea``.

    strip
        If true, during deserialization, strip the value of leading
        and trailing whitespace (default ``True``).
    """
    template = 'textarea'
    cols = 60
    rows = 10
    strip = True

class RichTextWidget(TextInputWidget):
    """
    Renders a ``<textarea>`` widget with the
    :term:`TinyMCE Editor`.

    To use this widget the :term:`TinyMCE Editor` library must be
    provided in the page where the widget is rendered. A version of
    :term:`TinyMCE Editor` is included in reform's static directory.
    
    **Configuration settings**
    
    These settings can be set in the fields configuration 
    (`nive.definitions.FieldConf`) as part of the settings. E.g.
    
        field.settings["options"] = {...} # Tiny mce options 
    
    width
        The width to be used with the default options. Ignored if 
        options are set.
    
    height
        The width to be used with the default options. Ignored if 
        options are set.
    
    options
        Tiny mce configuration options. See tiny mce docs for all
        possible values. If options is None the widgets defaults will
        be used. (`RichTextWidget.options`)
        
    **Attributes/Arguments**

    delayed_load
        If you have many richtext fields, you can set this option to
        ``true``, and the richtext editor will only be loaded when
        clicking on the field (default ``false``)

    strip
        If true, during deserialization, strip the value of leading
        and trailing whitespace (default ``True``).

    template
        The template name used to render the widget.  Default:
        ``richtext``.

    skin 
        The skin for the WYSIWYG editor. Normally only needed if you
        plan to reuse a TinyMCE js from another framework that
        defined a skin.

    theme
        The theme for the WYSIWYG editor, ``simple`` or ``advanced``.
        Defaults to ``simple``.

    """
    delayed_load = False
    strip = True
    template = 'richtext'
    skin = 'default'
    theme = 'simple'
    category = 'default' #full-width'
    requirements = ( ('trumbowyg', None), )
    options = {
        "mode" : "exact",
        "elements": "",  # element id is set in renderOptions() call
        "height": "240",
        "width": "100%",
        "theme": "advanced",
        "theme_advanced_resizing": True,
        "theme_advanced_toolbar_align": "left",
        "theme_advanced_toolbar_location": "top",
        "plugins": "table,contextmenu,paste,wordcount",
        "convert_urls": False,

        # Theme options
        "theme_advanced_buttons1" : "bold,italic,styleselect,justifyleft,justifycenter,justifyright,bullist,numlist,link,unlink,table,|,cut,copy,paste,pasteword,|,undo,redo,|,code",
        "theme_advanced_buttons2" : "",
        "theme_advanced_buttons3" : "",
        "theme_advanced_buttons4" : "",
        "theme_advanced_statusbar_location" : "bottom",

        # Style formats
        "style_formats" : [
            {"title" : _("Header 1"), "block" : "h1"},
            {"title" : _("Header 2"), "block" : "h2"},
            {"title" : _("Header 3"), "block" : "h3"},
            {"title" : _("Header 4"), "block" : "h4"},
            {"title" : _("Text (p)"), "block" : "p"},
            {"title" : _("Formatted (pre)"), "block" : "pre"},
        ],

        # Content CSS (should be your site CSS)
        #content_css : "",

        # Drop lists for link/image/media/template dialogs
        #external_link_list_url : "tiny_mce/lists/link_list.js",

        # Replace values for the template plugin
        #template_replace_values : {}   
    }
    
    def renderOptions(self, field):
        """
        Renders the tiny mce configuration options to be included in the template
        
        Custom Tiny MCE options can be specified in the field configuration
        (`nive.definitions.FieldConf`) as:
    
            field.settings["options"]        

        The options should either be a string or dictionary. Dicts will be rendered 
        as json strings.
        
        Width and height of the editor field can be customized for the default
        options by setting:
        
            field.settings["width"]
            field.settings["height"]

        These values will be ignored if `field.settings["options"]` set.
        """
        if self.configuration and self.configuration.get("settings"):
            opt = self.configuration.settings.get("options")
            if opt:
                if isinstance(opt, dict):
                    opt["elements"] = field.oid
                    return json.dumps(opt)
                return opt % {"oid": field.oid}

        # set editor width and height for default options
        opt = copy.deepcopy(self.options)
        opt["elements"] = field.oid
        if self.configuration and self.configuration.get("settings"):
            opt["width"] = self.configuration.settings.get("width") or opt["width"]
            opt["height"] = self.configuration.settings.get("height") or opt["height"]
        return json.dumps(opt)
        

class CodeTextWidget(TextInputWidget):
    """
    Renders a ``<textarea>`` widget with the
    :term:`Codemirror Editor`.

    **Attributes/Arguments**

    height
        The height, in pixels, of the text editor.  Defaults to 240.

    template
        The template name used to render the widget.  Default:
        ``codetext``.

    codetype 
        The skin for the WYSIWYG editor. Normally only needed if you
        plan to reuse a TinyMCE js from another framework that
        defined a skin.

    theme
        The theme for the WYSIWYG editor, ``simple`` or ``advanced``.
        Defaults to ``simple``.

    width
        The width, in pixels, of the editor.  Defaults to 500.
        The width can also be given as a percentage (e.g. '100%')
        relative to the width of the enclosing element.
    """
    height = 400
    width = 500
    template = 'codetext'
    codetype = 'default'
    category = 'default' #full-width'
    theme = 'simple'
    requirements = ( ('codemirror', None), )
    options = { "mode": "text/html", "tabMode": "indent", "lineNumbers": True }

    def renderOptions(self, field):
        """
        Renders the Codemirror configuration options to be included in the template
        
        Custom Codemirror options can be specified in the field configuration
        (`nive.definitions.FieldConf`) as:
    
            field.settings["options"]        

        The options should either be a string or dictionary. Dicts will be rendered 
        as json strings.
        
        Width and height of the editor field can be customized for the default
        options by setting:
        
            field.settings["width"]
            field.settings["height"]

        These values will be ignored if `field.settings["options"]` set.
        """
        if self.configuration and self.configuration.get("settings"):
            opt = self.configuration.settings.get("options")
            if opt:
                if isinstance(opt, dict):
                    return json.dumps(opt)
                return opt

        # set editor width and height for default options
        return json.dumps(self.options)

class PasswordWidget(TextInputWidget):
    """
    Renders a single <input type="password"/> input field.

    **Attributes/Arguments**

    template
        The template name used to render the widget.  Default:
        ``password``.

    size
        The ``size`` attribute of the password input field (default:
        ``None``).

    strip
        If true, during deserialization, strip the value of leading
        and trailing whitespace (default ``True``).
    """
    template = 'password'

class HiddenWidget(Widget):
    """
    Renders an ``<input type="hidden">`` widget.

    **Attributes/Arguments**

    template
        The template name used to render the widget.  Default:
        ``hidden``.
    """
    template = 'hidden'
    hidden = True

    def serialize(self, field, cstruct):
        if cstruct in (null, None):
            cstruct = ''
        return field.renderer(self.template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if not pstruct:
            return null
        return pstruct

class CheckboxWidget(Widget):
    """
    Renders an ``<input type="checkbox">`` widget.

    **Attributes/Arguments**

    true_val
        The value which should be returned during deserialization if
        the box is checked.  Default: ``true``.

    false_val
        The value which should be returned during deserialization if
        the box was left unchecked.  Default: ``false``.

    template
        The template name used to render the widget.  Default:
        ``checkbox``.

    """
    true_val = 'true'
    false_val = 'false'
    title = ''
    css_class = 'form-check-input'
    template = 'checkbox'

    def serialize(self, field, cstruct):
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct is null:
            return self.false_val
        return (pstruct == self.true_val) and self.true_val or self.false_val

class SelectWidget(Widget):
    """
    Renders ``<select>`` field based on a predefined set of values.

    **Attributes/Arguments**

    values
        A sequence of two-tuples (both values must be **string** or
        **unicode** values) indicating allowable, displayed values,
        e.g. ``( ('true', 'True'), ('false', 'False') )``.  The first
        element in the tuple is the value that should be returned when
        the form is posted.  The second is the display value.

    size
        The ``size`` attribute of the select input field (default:
        ``None``).

    null_value
        The value which represents the null value.  When the null
        value is encountered during serialization, the
        :attr:`colander.null` sentinel is returned to the caller.
        Default: ``''`` (the empty string).

    template
        The template name used to render the widget.  Default:
        ``select``.

    """
    template = 'select'
    null_value = None
    values = ()
    size = None

    def serialize(self, field, cstruct):
        if cstruct in (null, None):
            cstruct = self.null_value
        template = self.template
        if isinstance(cstruct, (list,tuple)):
            cstruct = [str(a) for a in cstruct]
        return field.renderer(template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct in (null, self.null_value):#, ""
            return null
        return pstruct
    
    def controlset_fields(self, field, value=None, format='html'):
        """
        Returns a list of controlset field ids. If `value` is None all controlled
        fields will be returned otherwise only the ids linked to this value.
        """
        items = self.form.GetField(field.name).get("listItems",[])
        if not isinstance(items, (list,tuple)):
            items = items(field, None)
        ids = []
        for i in items:
            # check list item options. For a control set each item may be a field.
            if value and value!=i.id:
                continue
            try:
                # it is a field list. add these ids too. 
                for v in i.fields:
                    ids.append(v.id)
            except AttributeError:
                pass
        if format=="html":
            return json.dumps(ids)
        return ids


class ChooseWidget(SelectWidget):
    """
    Extends SelectWidget for long select lists with search option
    """
    template = 'choose'
    requirements = ( ('chosen', None), )
    option_defaults = {}
    null_value = ''

    def serialize(self, field, cstruct):
        if cstruct in (null, None):
            cstruct = self.null_value
        template = self.template
        if isinstance(cstruct, (list, tuple)):
            cstruct = [str(a) for a in cstruct]
        return field.renderer(template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct in (null, None, ""):
            return self.null_value
        return pstruct


class UnitWidget(SelectWidget):
    """
    Extends SelectWidget for long select lists with search option
    """
    template = 'unitselect'
    requirements = ( ('chosen', None), )
    option_defaults = {}


class RadioChoiceWidget(SelectWidget):
    """
    Renders a sequence of ``<input type="radio">`` buttons based on a
    predefined set of values.

    **Attributes/Arguments**

    values
        A sequence of two-tuples (both values must be **string** or
        **unicode** values) indicating allowable, displayed values,
        e.g. ``( ('true', 'True'), ('false', 'False') )``.  The first
        element in the tuple is the value that should be returned when
        the form is posted.  The second is the display value.

    template
        The template name used to render the widget.  Default:
        ``radio_choice``.

    null_value
        The value used to replace the ``colander.null`` value when it
        is passed to the ``serialize`` or ``deserialize`` method.
        Default: the empty string.
    """
    template = 'radio_choice'
    css_class = 'form-check-input'


class CheckboxChoiceWidget(Widget):
    """
    Renders a sequence of ``<input type="check">`` buttons based on a
    predefined set of values.

    **Attributes/Arguments**

    values
        A sequence of two-tuples (both values must be **string** or
        **unicode** values) indicating allowable, displayed values,
        e.g. ``( ('true', 'True'), ('false', 'False') )``.  The first
        element in the tuple is the value that should be returned when
        the form is posted.  The second is the display value.

    template
        The template name used to render the widget.  Default:
        ``checkbox_choice``.

    null_value
        The value used to replace the ``colander.null`` value when it
        is passed to the ``serialize`` or ``deserialize`` method.
        Default: the empty string.
    """
    null_value = ()
    template = 'checkbox_choice'
    css_class = 'form-check-input'
    values = ()

    def serialize(self, field, cstruct):
        if cstruct in (null, None):
            cstruct = self.null_value
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct is null:
            return self.null_value
        if isinstance(pstruct, str):
            return (pstruct,)
        return tuple(pstruct)

class CheckedInputWidget(Widget):
    """
    Renders two text input fields: 'value' and 'confirm'.
    Validates that the 'value' value matches the 'confirm' value.

    **Attributes/Arguments**

    template
        The template name used to render the widget.  Default:
        ``checked_input``.

    size
        The ``size`` attribute of the input fields (default:
        ``None``, default browser size).

    mismatch_message
        The message to be displayed when the value in the primary
        field doesn't match the value in the confirm field.

    mask
        A :term:`jquery.maskedinput` input mask, as a string.  Both
        input fields will use this mask.

        a - Represents an alpha character (A-Z,a-z)
        9 - Represents a numeric character (0-9)
        * - Represents an alphanumeric character (A-Z,a-z,0-9)

        All other characters in the mask will be considered mask
        literals.

        Example masks:

          Date: 99/99/9999

          US Phone: (999) 999-9999

          US SSN: 999-99-9999

        When this option is used, the :term:`jquery.maskedinput`
        library must be loaded into the page serving the form for the
        mask argument to have any effect.  See :ref:`masked_input`.

    mask_placeholder
        The placeholder for required nonliteral elements when a mask
        is used.  Default: ``_`` (underscore).
    """
    template = 'checked_input'
    size = None
    mismatch_message = _('Fields did not match')
    subject = _('Value')
    confirm_subject = _('Confirm Value')
    mask = None
    mask_placeholder = "_"
    requirements = ( ('jquery.maskedinput', None), )

    def serialize(self, field, cstruct):
        if cstruct in (null, None):
            cstruct = ''
        confirm = getattr(field, '%s-confirm' % (field.name,), '')
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct,
                              confirm=confirm, subject=self.subject,
                              confirm_subject=self.confirm_subject,
                              )

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct is null:
            return null
        value = pstruct
        confirm = ''
        if formstruct:
            confirm = formstruct.get('%s-confirm' % (field.name,)) or ''
        setattr(field, '%s-confirm' % (field.name,), confirm)
        if (value or confirm) and (value != confirm):
            raise Invalid(field.schema, self.mismatch_message, value)
        if not value:
            return null
        return value

class CheckedPasswordWidget(CheckedInputWidget):
    """
    Renders two password input fields: 'password' and 'confirm'.
    Validates that the 'password' value matches the 'confirm' value.

    **Attributes/Arguments**

    template
        The template name used to render the widget.  Default:
        ``password_confirm``.

    size
        The ``size`` attribute of the password input field (default:
        ``None``).
    
    update
        The ``update`` attribute prevents rendering of existing passwords.
        Instead ``placeholder`` is used.
        
    """
    template = 'password_confirm'
    mismatch_message = _('Password did not match confirm')
    size = None
    update = False
    placeholder = ''

    def serialize(self, field, cstruct):
        if self.update and cstruct in (null, None, ''):
            cstruct = confirm = self.placeholder
        elif cstruct in (null, None):
            cstruct = confirm = ''
        else:
            confirm = getattr(field, '%s-confirm' % (field.name,), '')
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct,
                              confirm=confirm, subject=self.subject,
                              confirm_subject=self.confirm_subject,
                              )

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct in (null, self.placeholder):
            return null
        value = pstruct
        confirm = formstruct.get('%s-confirm' % (field.name,)) or ''
        setattr(field, '%s-confirm' % (field.name,), confirm)
        if (value or confirm) and (value != confirm):
            raise Invalid(field.schema, self.mismatch_message, value)
        if not value:
            return null
        return value




#<removed class MappingWidget(Widget):
#<removed class SequenceWidget(Widget):

class FileUploadWidget(Widget):
    """
    Represent a file upload.  Meant to work with a
    :class:`deform.FileData` schema node.

    This widget accepts a single required positional argument in its
    constructor: ``tmpstore``.  This argument should be passed an
    instance of an object that implements the
    :class:`deform.interfaces.FileUploadTempStore` interface.  Such an
    instance will hold on to file upload data during the validation
    process, so the user doesn't need to reupload files if other parts
    of the form rendering fail validation.  See also
    :class:`deform.interfaces.FileUploadTempStore`.

    **Attributes/Arguments**

    template
        The template name used to render the widget.  Default:
        ``file_upload``.

    readonly_template
        The template name used to render the widget in read-only mode.
        Default: ``readonly/file_upload``.

    size
        The ``size`` attribute of the input field (default ``None``).
    """
    template = 'file_upload'
    readonly_template = 'readonly/file_upload'
    size = None

    def __init__(self, tmpstore, **kw):
        Widget.__init__(self, **kw)
        self.tmpstore = tmpstore

    def random_id(self):
        return ''.join(
            [random.choice('ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789') for i in range(10)])

    def serialize(self, field, cstruct, readonly=False):
        if cstruct in (null, None):
            cstruct = {}
        if cstruct:
            uid = cstruct['uid']
            if not uid in self.tmpstore:
                self.tmpstore[uid] = cstruct

        template = readonly and self.readonly_template or self.template
        return field.renderer(template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct is null:
            return null
        # switch to formstruct ?
        upload = pstruct.get('upload')
        uid = pstruct.get('uid')

        if hasattr(upload, 'file'):
            # the upload control had a file selected
            data = schema.filedict()
            data['fp'] = upload.file
            data['filename'] = ntpath.basename(upload.filename)
            data['mimetype'] = upload.type
            data['size']  = upload.length
            if uid is None:
                # no previous file exists
                while 1:
                    uid = self.random_id()
                    if self.tmpstore.get(uid) is None:
                        data['uid'] = uid
                        data['preview_url'] = self.tmpstore.preview_url(uid)
                        self.tmpstore[uid] = data
                        break
            else:
                # a previous file exists
                data['uid'] = uid
                data['preview_url'] = self.tmpstore.preview_url(uid)
                self.tmpstore[uid] = data
        else:
            # the upload control had no file selected
            if uid is None:
                # no previous file exists
                return null
            else:
                # a previous file should exist
                data = self.tmpstore.get(uid)
                # but if it doesn't, don't blow up
                if data is None:
                    return null

        return data


class FileUploadWidget2(Widget):
    """
    Represent a file upload.  Meant to work with a
    :class:`reform.FileData2` schema node.

    Extended version using nive `File` class instead of
    schema.filedict().
     
    **Attributes/Arguments**

    template
        The template name used to render the widget.  Default:
        ``file_upload``.

    size
        The ``size`` attribute of the input field (default ``None``).
    """
    template = 'file_upload'
    size = None

    def __init__(self, **kw):
        Widget.__init__(self, **kw)

    def serialize(self, field, cstruct):
        if cstruct in ("", b"", null, None):
            cstruct = {}
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct in ("", b"", null, None):
            return null
        try:
            cls = self.form.app.db.GetFileClass()
        except AttributeError:
            cls = File
        file = cls(
            filename = ntpath.basename(pstruct.filename),
            file = pstruct.file,
            filekey = field.name,
            size = pstruct.length if hasattr(pstruct, "length") else -1,
            tempfile = True
        )
        # check file size, might be -1
        if 0 and file.size==-1 and file.file:
            file.file.seek(0, os.SEEK_END)
            size = file.file.tell()
            # reset
            file.file.seek(0)
            if size:
                file.size = size
        if hasattr(pstruct, "type"):
            file.mime = pstruct.type
        return file


class FileToDataUploadWidget(Widget):
    """
    Represent a file upload. Uploaded file data is extracted and
    return as simple data/string object. Optionally data can be encoded
    as base64 to get a valid string.

    Extended version using nive `File` class instead of
    schema.filedict().

    **Attributes/Arguments**

    template
        The template name used to render the widget.  Default:
        ``filetodata_upload``.

    size
        The ``size`` attribute of the input field (default ``None``).
    """
    template = 'filetodata_upload'
    size = None
    base64 = False

    def __init__(self, **kw):
        Widget.__init__(self, **kw)

    def serialize(self, field, cstruct):
        if cstruct in ("", null, None):
            cstruct = {}
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct)

    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct in ("", null, None):
            # check if delfile checkbox ticked and return an empty string
            # to reset field contents
            if formstruct and formstruct.get(field.name+"_delfile")=="1":
                return ""
            return null
        file = pstruct.file.read()
        if self.base64:
            file = base64.b64encode(bytes(file,encoding="utf-8")) # todo [3] unicode
        return file




class DatePartsWidget(Widget):
    """
    Renders a set of ``<input type='text'>`` controls based on the
    year, month, and day parts of the serialization of a
    :class:`colander.Date` object or a string in the format
    ``YYYY-MM-DD``.  This widget is usually meant to be used as widget
    which renders a :class:`colander.Date` type; validation
    likely won't work as you expect if you use it against a
    :class:`colander.String` object, but it is possible to use it
    with one if you use a proper validator.

    **Attributes/Arguments**

    template
        The template name used to render the input widget.  Default:
        ``dateparts``.

    size
        The size (in columns) of each date part input control.
        Default: ``None`` (let browser decide).

    assume_y2k
        If a year is provided in 2-digit form, assume it means
        2000+year.  Default: ``True``.

    """
    template = 'dateparts'
    size = None
    assume_y2k = True

    def serialize(self, field, cstruct):
        if cstruct is null:
            year = ''
            month = ''
            day = ''
        else:
            year, month, day = cstruct.split('-', 2)
        template = self.template
        return field.renderer(template, field=field, cstruct=cstruct,
                              year=year, month=month, day=day)

    def deserialize(self, field, pstruct, formstruct=None):
        if formstruct and 'year' in formstruct and 'month' in formstruct and 'day' in formstruct:
            year = formstruct['year'].strip()
            month = formstruct['month'].strip()
            day = formstruct['day'].strip()
            
            if (not year and not month and not day):
                return null

            if self.assume_y2k and len(year) == 2:
                year = '20' + year
            result = '-'.join([year, month, day])

            if (not year or not month or not day):
                raise Invalid(field.schema, _('Incomplete date'), result)
        elif pstruct is null:
            return null
        else:
            result = pstruct

        return result

class TextAreaCSVWidget(Widget):
    """
    Widget used for a sequence of tuples of scalars; allows for
    editing CSV within a text area.  Used with a schema node which is
    a sequence of tuples.
    
    **Attributes/Arguments**

    cols
        The size, in columns, of the text input field.  Defaults to
        ``None``, meaning that the ``cols`` is not included in the
        widget output (uses browser default cols).

    rows
        The size, in rows, of the text input field.  Defaults to
        ``None``, meaning that the ``rows`` is not included in the
        widget output (uses browser default cols).

    template
        The template name used to render the widget.  Default:
        ``textarea``.

    """
    template = 'textarea'
    cols = None
    rows = None

    def serialize(self, field, cstruct):
        if cstruct is null:
            cstruct = []
        textrows = getattr(field, 'unparseable', None)
        if textrows is None:
            outfile = io.StringIO()
            writer = csv.writer(outfile)
            writer.writerows(cstruct)
            textrows = outfile.getvalue()
        template = self.template
        return field.renderer(template, field=field, cstruct=textrows)
        
    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct is null:
            return null
        if not pstruct.strip():
            return null
        try:
            infile = io.StringIO(pstruct)
            reader = csv.reader(infile)
            rows = list(reader)
        except Exception as e:
            field.unparseable = pstruct
            raise Invalid(field.schema, str(e))
        return rows

    def handle_error(self, field, error):
        msgs = []
        if error.msg:
            field.error = error
        else:
            for e in error.children:
                msgs.append('line %s: %s' % (e.pos+1, e))
            field.error = Invalid(field.schema, '\n'.join(msgs))


class TextInputCSVWidget(Widget):
    """
    Widget used for a tuple of scalars; allows for editing a single
    CSV line within a text input.  Used with a schema node which is a
    tuple composed entirely of scalar values (integers, strings, etc).
    
    **Attributes/Arguments**

    template
        The template name used to render the widget.  Default:
        ``textinput``.

    size
        The size, in columns, of the text input field.  Defaults to
        ``None``, meaning that the ``size`` is not included in the
        widget output (uses browser default size).
    """
    template = 'textinput'
    size = None
    mask = None

    def serialize(self, field, cstruct):
        if cstruct is null:
            cstruct = ''
        textrow = getattr(field, 'unparseable', None)
        if textrow is None:
            outfile = io.StringIO()
            writer = csv.writer(outfile)
            writer.writerow(cstruct)
            textrow = outfile.getvalue().strip()
        template = self.template
        return field.renderer(template, field=field, cstruct=textrow)
        
    def deserialize(self, field, pstruct, formstruct=None):
        if pstruct is null:
            return null
        if not pstruct.strip():
            return null
        try:
            infile = io.StringIO(pstruct)
            reader = csv.reader(infile)
            row = next(reader)
        except Exception as e:
            field.unparseable = pstruct
            raise Invalid(field.schema, str(e))
        return row

    def handle_error(self, field, error):
        msgs = []
        if error.msg:
            field.error = error
        else:
            for e in error.children:
                msgs.append('%s' % e)
            field.error = Invalid(field.schema, '\n'.join(msgs))


class ResourceRegistry(object):
    """ A resource registry maps :term:`widget requirement` name/version
    pairs to one or more relative resources.  A resource registry can
    be passed to a :class:`reform.Form` constructor; if a resource
    registry is *not* passed to the form constructor, a default
    resource registry is used by that form.  The default resource
    registry contains only mappings from requirement names to
    resources required by the built-in Deform widgets (not by any
    add-on widgets).

    If the ``use_defaults`` flag is True, the default set of Deform
    requirement-to-resource mappings is loaded into the registry.
    Otherwise, the registry is initialized without any mappings.
    """
    def __init__(self, use_defaults=True):
        if use_defaults is True:
            self.registry = default_resources.copy()
        else:
            self.registry = {}

    def set_js_resources(self, requirement, version, *resources):
        """ Set the Javascript resources for the requirement/version
        pair, using ``resources`` as the set of relative resource paths."""
        reqt = self.registry.setdefault(requirement, {})
        ver = reqt.setdefault(version, {})
        ver['js'] = resources

    def set_css_resources(self, requirement, version, *resources):
        """ Set the CSS resources for the requirement/version
        pair, using ``resources`` as the set of relative resource paths."""
        reqt = self.registry.setdefault(requirement, {})
        ver = reqt.setdefault(version, {})
        ver['css'] = resources

    def __call__(self, requirements):
        """ Return a dictionary representing the resources required
        for a particular set of requirements (as returned by
        :meth:`reform.Field.get_widget_requirements`).  The dictionary
        will be a mapping from resource type (``js`` and ``css`` are
        both keys in the dictionary) to a list of relative resource
        paths.  Each path is relative to wherever yo've mounted
        Deform's ``static`` directory in your web server.  You can use
        the paths for each resource type to inject CSS and Javascript
        on-demand into the head of dynamic pages that render Deform
        forms.  """
        result = {'js':[], 'css':[], 'seq':[]}
        for requirement, version in requirements:
            tmp = self.registry.get(requirement)
            if tmp is None:
                raise ValueError(
                    'Cannot resolve widget requirement %r' % requirement)
            versioned = tmp.get(version)
            if versioned is None:
                raise ValueError(
                    'Cannot resolve widget requirement %r (version %r)' % (
                        (requirement, version)))
            for thing in ('js', 'css', 'seq'):
                sources = versioned.get(thing)
                if sources is None:
                    continue
                if not hasattr(sources, '__iter__'):
                    sources = (sources,)
                for source in sources:
                    if not source in result[thing]:
                        result[thing].append(source)
        return result

            
default_resources = {
    'jquery': {
        None:{
            'seq':(('jquery.js', 'nive.components.reform:static/scripts/jquery.min.js'),),
            },
        },
    'jqueryui': {
        None:{
            'seq':(('jquery.js', 'nive.components.reform:static/scripts/jquery.min.js'),
                   ('jquery-ui.js', 'nive.components.reform:static/scripts/jquery-ui-1.8.11.custom.min.js'),
                   ('jquery-ui.css', 'nive.components.reform:static/css/ui-lightness/jquery-ui-1.8.11.custom.css'))
            },
        },
    'deactivated!jquery.form': {
        None:{
            'seq':(('jquery.js', 'nive.components.reform:static/scripts/jquery.min.js'),
                   ('jquery.form.js', 'nive.components.reform:static/scripts/jquery.form-3.50.js')),
            },
        },
    'jquery.maskedinput': {
        None:{
            'seq':(('jquery.js', 'nive.components.reform:static/scripts/jquery.min.js'),
                   ('jquery.maskedinput.js', 'nive.components.reform:static/scripts/jquery.maskedinput-1.3.1.min.js')),
            },
        },
    'datetimepicker': {
        None:{
            'seq':(('jquery.js', 'nive.components.reform:static/scripts/jquery.min.js'),
                   ('moment.js', 'nive.components.reform:static/tempusdominus/moment.min.js'),
                   ('moment-de.js', 'nive.components.reform:static/tempusdominus/de.js'),
                   ('tempusdominus.js', 'nive.components.reform:static/tempusdominus/tempusdominus-bootstrap-4.min.js'),
                   ('tempusdominus.css', 'nive.components.reform:static/tempusdominus/tempusdominus-bootstrap-4.min.css')),
            },
        },
    'reform': {
        None:{
            'seq':(('jquery.js', 'nive.components.reform:static/scripts/jquery.min.js'),
                   #('jquery.form.js', 'nive.components.reform:static/scripts/jquery.form.js'),
                   ('reform.js', 'nive.components.reform:static/scripts/reform.js')),
            },
        },
    'trumbowyg': {
        None:{
            'seq':(('trumbowyg.js', 'nive.components.reform:static/trumbowyg/trumbowyg.min.js'),
                   ('trumbowyg.cleanpaste.js', 'nive.components.reform:static/trumbowyg/plugins/cleanpaste/trumbowyg.cleanpaste.js'),
                   ('trumbowyg.de.js', 'nive.components.reform:static/trumbowyg/langs/de.min.js'),
                   ('trumbowyg.css', 'nive.components.reform:static/trumbowyg/ui/trumbowyg.min.css')),
            },
        },
    'chosen': {
        None:{
            'seq':(('chosen.js', 'nive.components.reform:static/chosen/chosen.jquery.min.js'),
                   ('chosen.css', 'nive.components.reform:static/chosen/chosen.min.css')),
            },
        },
    'codemirror': {
        None:{
            'seq':(('codemirror.js', 'nive.components.reform:static/codemirror/lib/codemirror.js'),
                   ('codemirror-xml.js', 'nive.components.reform:static/codemirror/mode/xml/xml.js'),
                   ('codemirror-javascript.js', 'nive.components.reform:static/codemirror/mode/javascript/javascript.js'),
                   ('codemirror-css.js', 'nive.components.reform:static/codemirror/mode/css/css.js'),
                   ('codemirror-htmlmixed.js', 'nive.components.reform:static/codemirror/mode/htmlmixed/htmlmixed.js'),
                   ('codemirror.css', 'nive.components.reform:static/codemirror/lib/codemirror.css')),
            },
        },
    }

default_resource_registry = ResourceRegistry()

default_widget_makers = {
    #<removed schema.Mapping: widget.MappingWidget,
    #<removed schema.Sequence: widget.SequenceWidget,
    schema.String: TextInputWidget,
    schema.Integer: TextInputWidget,
    schema.Float: TextInputWidget,
    schema.Decimal: TextInputWidget,
    schema.Boolean: CheckboxWidget,
    schema.Date: DateInputWidget,
    schema.DateTime: DateTimeInputWidget,
    schema.Tuple: TextInputCSVWidget,
    schema.Set: CheckboxChoiceWidget,
    schema.List: CheckboxChoiceWidget
}